/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.ConvolveOp;
import java.awt.image.Kernel;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;

public class BlurringKerning4 implements Kerning<BlurringKerning4.DoubleOutline> {

    private int maximalOverlapArea;
    private int blurSmallSize;
    private int blurSmallThreshold;
    private int blurBigSize;
    private int blurBigThreshold;

    static public class Outline {
        public final int[] x1s, x2s;
        public final int width;

        public Outline(int[] x1s, int[] x2s, int width) {
            this.x1s = x1s;
            this.x2s = x2s;
            this.width = width;
        }
    }

    static public class DoubleOutline {
        public final boolean symbol;
        public final Outline softBig, hardSmall;

        public DoubleOutline(Outline softBig, Outline hardSmall, boolean symbol) {
            this.softBig = softBig;
            this.hardSmall = hardSmall;
            this.symbol = symbol;
        }
    }

    public BlurringKerning4(int maximalOverlapArea, int blurSmallSize, int blurSmallThreshold, int blurBigSize, int blurBigThreshold) {
        this.maximalOverlapArea = maximalOverlapArea;
        this.blurSmallSize = blurSmallSize;
        this.blurSmallThreshold = blurSmallThreshold;
        this.blurBigSize = blurBigSize;
        this.blurBigThreshold = blurBigThreshold;
    }

    @Override
    public DoubleOutline kernPreProcess(BufferedImage bi, String glyphName) {
        BufferedImage small = getBlurredImage(bi, blurSmallSize);
        BufferedImage smallTh = thresholdImage(small, blurSmallThreshold);
//        saveImage(smallTh, "kern-" + glyphName + "-small-si" + blurSmallSize + "-th" + blurSmallThreshold);
        BufferedImage big = getBlurredImage(bi, blurBigSize);
        BufferedImage bigTh = thresholdImage(big, blurBigThreshold);
//        saveImage(bigTh, "kern-" + glyphName + "-big-si" + blurBigSize + "-th" + blurBigThreshold);
        Outline outlineHardSmall = getOutline(smallTh);
        Outline outlineSoftBig = getOutline(bigTh);
        return new DoubleOutline(outlineSoftBig, outlineHardSmall, glyphName.startsWith("_") && !glyphName.startsWith("__"));
    }

    private static BufferedImage thresholdImage(BufferedImage bi, int threshold) {
        WritableRaster r = bi.getRaster();
        int width = r.getWidth();
        int height = r.getHeight();
        BufferedImage resultBI = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        WritableRaster resultR = resultBI.getRaster();
        int[] black = new int[]{0};
        int[] white = new int[]{255};
        int[] iArray = new int[4];
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                r.getPixel(x, y, iArray);
                resultR.setPixel(x, y, iArray[3] > threshold ? black : white);
            }
        }
        return resultBI;
    }

    private static BufferedImage getBlurredImage(BufferedImage bi, int n) {
        float f = 1f / n;
        float[] floats = new float[n];
        for (int idx = 0; idx < n; idx++) {
            floats[idx] = f;
        }
        Kernel kernel1 = new Kernel(1, n, floats);
        Kernel kernel2 = new Kernel(n, 1, floats);
        BufferedImageOp op1 = new ConvolveOp(kernel1, ConvolveOp.EDGE_ZERO_FILL, null);
        BufferedImageOp op2 = new ConvolveOp(kernel2, ConvolveOp.EDGE_ZERO_FILL, null);
        BufferedImage bi2 = op1.filter(bi, null);
        BufferedImage bi3 = op2.filter(bi2, null);
        op1.filter(bi3, bi2);
        op2.filter(bi2, bi3);
        op1.filter(bi3, bi2);
        op2.filter(bi2, bi3);
        return bi3;
    }

    private static Outline getOutline(BufferedImage bi) {
        WritableRaster r = bi.getRaster();
        int width = r.getWidth();
        int height = r.getHeight();
        int[] iArray = new int[1];
        int[] x1s = new int[height];
        int[] x2s = new int[height];
        for (int y = 0; y < height; y++) {
            int minx = width, maxx = -1;
            for (int x = 0; x < width; x++) {
                r.getPixel(x, y, iArray);
                if (iArray[0] == 0) {
                    if (x < minx) {
                        minx = x;
                    }
                    if (x > maxx) {
                        maxx = x;
                    }
                }
            }
            x1s[y] = minx;
            x2s[y] = maxx;
        }
        return new Outline(x1s, x2s, width);
    }

    private static void saveImage(BufferedImage bi, String name) {
        try {
            ImageIO.write(bi, "png", new File(name + ".png"));
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public int kernEval(DoubleOutline left, DoubleOutline right) {
        int minimal = getDistance(left.hardSmall.x2s, right.hardSmall.x1s, right.hardSmall.width, left.symbol, right.symbol);
        int dist = getDistanceOverlapLimit(minimal, left.softBig.x2s, right.softBig.x1s, right.softBig.width, maximalOverlapArea);
        return dist;
    }

    private static int getDistanceOverlapLimit(int minimal, int[] l2s, int[] r1s, int width, int maximalOverlapArea) {
        for (int xshift = width; xshift >= minimal; xshift--) {
            int overlapArea = 0;
            for (int y = 0; y < l2s.length; y++) {
                int l2 = l2s[y];
                int r1 = r1s[y];
                int overlap = Math.max(0, l2 - (r1 + xshift));
                overlapArea += overlap;
            }
            if (overlapArea > maximalOverlapArea) {
                return xshift;
            }
        }
        return minimal;
    }

    private static int getDistance(int[] l2s, int[] r1s, int w2, boolean lsymbol, boolean rsymbol) {
        int max = l2s[0] - r1s[0];
        boolean met = false;
        if (!lsymbol) { // kern alphanum-alphanum, alphanum-symbol, but not symbol-alphanum nor symbol-symbol
            for (int y = 0; y < l2s.length; y++) {
                if (l2s[y] != -1 && r1s[y] != w2) {
                    met = true;
                }
                int dist = l2s[y] - r1s[y];
                if (dist > max) {
                    max = dist;
                }
            }
        }
        if (met) {
            return max;
        } else {
            // use bounding box
            int l2sMax = l2s[0];
            for (int l2 : l2s) {
                l2sMax = Math.max(l2sMax, l2);
            }
            int r1sMin = r1s[0];
            for (int r1 : r1s) {
                r1sMin = Math.min(r1sMin, r1);
            }
            return l2sMax - r1sMin;
        }
    }
}
