/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import cz.jtra.math.quine.designer.numbercompositor.NumberCompositor;

import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.filechooser.FileFilter;
import java.awt.*;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.*;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Compositor extends JFrame {
    public static final String FILE_EXT_COMPOSITION = ".composition";
    public static final String FILE_EXT_COMPOSITION_DESCRIPTION = "Glyph Composition files";
    public static final String FILE_EXT_NUMBER = ".number";
    public static final String FILE_EXT_GLYPH_LOCATIONS = ".glyphlocations";
    public static final Pattern PATTERN_NONWS_NONWS = Pattern.compile("(\\S\\S)|(\\A\\S)|(\\S\\z)");
    public static final Pattern PATTERN_WS_PLUS = Pattern.compile("(\\s+)");
    public static final Pattern PATTERN_POSITION = Pattern.compile("(([-+]?[0-9]+\\.[0-9]+)|(\\([-+]?[0-9]+\\.[0-9]+\\)))(x|y|ax|ay)?");
    public static final Pattern PATTERN_POSITION_SIMPLE_X = Pattern.compile("([-+]?[0-9]+\\.[0-9]+)");
    public static final Pattern PATTERN_GLYPH_ANCHOR = Pattern.compile("\\A([-+]?[0-9]+\\.[0-9]+),([-+]?[0-9]+\\.[0-9]+)an\\z");
    public static final Pattern PATTERN_SCALE_DIGIT = Pattern.compile("\\A([0-9])s\\z");
    private static final Pattern PATTERN_SCALE = Pattern.compile("scale:([0-9.]+)");
    private static final Pattern PATTERN_VIEW = Pattern.compile("view:([+-]?[0-9.]+),([+-]?[0-9.]+)");
    private static final Pattern PATTERN_GLYPH_MAPPING = Pattern.compile("([-a-zA-Z0-9_]+)=([-a-zA-Z0-9_.]+)");
    private static final Pattern PATTERN_KERN_SPACE_AS = Pattern.compile("space:([-a-zA-Z0-9_]+)");
    public static final Pattern PATTERN_TOKEN_DELIMITER = Pattern.compile("(\\s+)");
    public static final String GLYPH_SPACE = "_";

    private final JTextArea textField;
    private final JPanel renderPanel;
    private String lastText;
    private final StateManager stateManager;
    private final JTextField nameTextField;
    private int glyphPixels = 1000;
    private double glyphScale = 1.0d;
    private double glyphViewX = 0.0;
    private double glyphViewY = 0.0;
    private double glyphAnchorX = 0.2d;
    private double glyphAnchorY = 0.3d;
    private ArrayList<GlyphLocation> glyphLocations = new ArrayList<>();
    private SourceLocationInfo sourceLocationInfo;
    private Map<String, GlyphBitmap> glyphMapping = new HashMap<>();
    private Map<String, Object> kernPreProcessCache = new HashMap<>();
    private BufferedImage renderedTextBitmap;

    public Compositor(StateManager stateManager) {
        this.stateManager = stateManager;
        setBounds(new Rectangle(1200, 800));
        setTitle("Compositor");
        setLayout(new BorderLayout());
        JPanel bottomPanel = new JPanel();
        bottomPanel.setMinimumSize(new Dimension(300, 200));
        bottomPanel.setPreferredSize(new Dimension(300, 200));
        textField = new JTextArea();
        textField.setMargin(new Insets(4, 4, 4, 4));
        textField.addCaretListener(new CaretListener() {
            @Override
            public void caretUpdate(CaretEvent e) {
                updateOnTextFieldEvent();
            }
        });
        textField.addKeyListener(new KeyListener() {
            @Override
            public void keyTyped(KeyEvent e) {
                updateOnTextFieldEvent();
            }

            @Override
            public void keyPressed(KeyEvent e) {
            }

            @Override
            public void keyReleased(KeyEvent e) {
            }
        });
        bottomPanel.setLayout(new BorderLayout());
        JPanel separator = new JPanel();
        separator.setBackground(Color.darkGray);
        separator.setPreferredSize(new Dimension(100, 3));
        bottomPanel.add(separator, BorderLayout.NORTH);
        bottomPanel.add(new JScrollPane(textField, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED), BorderLayout.CENTER);
        JButton reloadButton = new JButton("Reload");
        reloadButton.addActionListener(e -> refreshAction());
        bottomPanel.add(reloadButton, BorderLayout.SOUTH);
        JPanel checkBoxPanel = new JPanel();
        checkBoxPanel.setLayout(new FlowLayout());
        JCheckBox highQualityCheckBox = new JCheckBox("High quality");
        highQualityCheckBox.addActionListener(e -> updateOnQualitySettingEvent());
        checkBoxPanel.add(highQualityCheckBox);
        JButton exportViewButton = new JButton("Export view");
        exportViewButton.addActionListener(e -> exportView());
        checkBoxPanel.add(exportViewButton);
        checkBoxPanel.add(new JLabel("Composition name:"));
        nameTextField = new JTextField(20);
        checkBoxPanel.add(nameTextField);
        JButton saveButton = new JButton("Save");
        saveButton.addActionListener(e -> save());
        checkBoxPanel.add(saveButton);
        JButton loadButton = new JButton("Load");
        loadButton.addActionListener(e -> load());
        checkBoxPanel.add(loadButton);
        JButton openButton = new JButton("Open");
        openButton.addActionListener(e -> open());
        checkBoxPanel.add(openButton);
        JButton opsButton = new JButton("Operations...");
        JPopupMenu opsMenu = new JPopupMenu();
        JMenuItem selectionMenu = new JMenuItem("Operations on selection:");
        selectionMenu.setEnabled(false);
        opsMenu.add(selectionMenu);
        JMenuItem rekern = new JMenuItem("Kern tokens");
        rekern.addActionListener(e -> modifySelection(this::kernTokens, true));
        opsMenu.add(rekern);
        JMenuItem kernWords = new JMenuItem("Split text to tokens");
        kernWords.addActionListener(e -> modifySelection(this::textToTokens, false));
        opsMenu.add(kernWords);
        opsMenu.addSeparator();
        JMenuItem setupMenu = new JMenuItem("Setup operations:");
        setupMenu.setEnabled(false);
        opsMenu.add(setupMenu);
        JMenuItem setupGlyphs = new JMenuItem("Setup latest glyphs");
        setupGlyphs.addActionListener(e -> modifySelection(this::setupGlyphs, false));
        opsMenu.add(setupGlyphs);
        JMenuItem demoGlyphs = new JMenuItem("Demo");
        demoGlyphs.addActionListener(e -> modifySelection(this::demoGlyphs, false));
        opsMenu.add(demoGlyphs);

        opsButton.addActionListener(e -> opsMenu.show(opsButton, 10, 10));
        checkBoxPanel.add(opsButton);
        bottomPanel.add(checkBoxPanel, BorderLayout.NORTH);
        renderPanel = new JPanel() {
            @Override
            protected void paintComponent(Graphics g) {
                System.out.println("paintComponent");
                boolean highQuality = highQualityCheckBox.isSelected();
                super.paintComponent(g);
                if (renderedTextBitmap != null && renderedTextBitmap.getWidth() == getWidth() && renderedTextBitmap.getHeight() == getHeight()) {
                    System.out.println("using cached text bitmap");
                } else {
                    System.out.println("drawing new text bitmap");
                    renderedTextBitmap = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_3BYTE_BGR);
                    paintGlyphs(renderedTextBitmap.getGraphics(), renderedTextBitmap.getWidth(), renderedTextBitmap.getHeight(), highQuality, glyphScale);
                }
                g.drawImage(renderedTextBitmap, 0, 0, null);
                paintCursor(g, getWidth(), getHeight(), glyphScale);//todo: glyphScale is passed as argument, but glyphViewX/Y not
                System.out.println("paint done");
            }
        };
        JSplitPane jSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        add(jSplitPane);
        jSplitPane.setResizeWeight(1);
        jSplitPane.setTopComponent(renderPanel);
        jSplitPane.setBottomComponent(bottomPanel);
    }

    private void updateOnTextFieldEvent() {
        String currentText = textField.getText();
        if (!currentText.equals(lastText)) {
            lastText = currentText;
            reloadOrRefresh(false);
        }
        renderPanel.repaint();
    }

    private void updateOnQualitySettingEvent() {
        reloadOrRefresh(false);
        renderPanel.repaint();
    }

    private String demoGlyphs(String before, String text, String after) {
        return demoOrSetupGlyphs(false) + cleanGlyphDefinitions(text);
    }


    private String setupGlyphs(String before, String text, String after) {
        return demoOrSetupGlyphs(true) + cleanGlyphDefinitions(text);
    }

    private String cleanGlyphDefinitions(String s) {
        ArrayList<String> tokenize = tokenize(s);
        tokenize.removeIf(tokenOrSpaces -> PATTERN_GLYPH_MAPPING.matcher(tokenOrSpaces).matches());
        return String.join("", tokenize);
    }

    public static final Pattern PATTERN_FILE_NUMBER_SUFFIX = Pattern.compile("(.*[^0-9])([0-9]+)([.]glyph)");

    private String demoOrSetupGlyphs(boolean onlySetup) {
        String directory = stateManager.getDirectory();
        File dir = new File(directory);
        File[] files = dir.listFiles(f -> f.isFile() && f.getName().endsWith(StateManager.FILE_EXT_GLYPH));
        Map<String, String> prefixToSuffix = new TreeMap<>();
        for (File file : files) {
            String name = file.getName();
            Matcher m = PATTERN_FILE_NUMBER_SUFFIX.matcher(name);
            if (m.matches()) {
                String prefix = m.group(1);
                String suffix = m.group(2);
                String oldSuffix = prefixToSuffix.get(prefix);
                if (oldSuffix == null || Integer.parseInt(oldSuffix) < Integer.parseInt(suffix)) {
                    prefixToSuffix.put(prefix, suffix);
                }
            }
        }
        StringBuilder sbDeclare = new StringBuilder();
        StringBuilder sbUse = new StringBuilder();
        for (Map.Entry<String, String> entry : prefixToSuffix.entrySet()) {
            String glyphName = entry.getKey();
            if (glyphName.startsWith("__") && glyphName.length() >= 3) {
                glyphName = glyphName.substring(2, 3).toUpperCase(Locale.US) + glyphName.substring(3);
            }
            sbDeclare.append(glyphName);
            sbDeclare.append('=');
            sbDeclare.append(entry.getKey());
            sbDeclare.append(entry.getValue());
            sbDeclare.append(' ');
            if (!onlySetup) {
                sbUse.append(glyphName);
                sbUse.append(' ');
                sbUse.append("0.4");
                sbUse.append(' ');
            }
        }
        if (onlySetup) {
            return sbDeclare.toString() + "\nspace:l\n\n";
        } else {
            return sbDeclare + "\nspace:l\n\n" + sbUse + "\n\nscale:0.2\nview:0.0,0.0\n";
        }
    }

    @FunctionalInterface
    interface ListChunkModify<T> {
        void chunkModify(ArrayList<T> items);
    }

    public static <T> ArrayList<T> listMapChunksOfN(int maxLength, int minLength, List<T> source, ListChunkModify<T> listModify) {
        ArrayList<T> result = new ArrayList<>(source);
        for (int i = 0; i <= result.size() - minLength; i++) {
            List<T> subList = result.subList(i, Math.min(i + maxLength, result.size()));
            if (subList.size() < minLength) {
                break;
            }
            ArrayList<T> modified = new ArrayList<>(subList);
            listModify.chunkModify(modified);
            subList.clear();
            result.addAll(i, modified);
        }
        return result;
    }

    private String kernTokens(String before, String text, String after) {
        String[] kernSpaceAs = new String[]{""};
        scanTokens(tokenize(before), PATTERN_KERN_SPACE_AS, m -> {
            kernSpaceAs[0] = m.group(1);
        });
        ArrayList<String> tokens = tokenize(text);
        if (tokens.size() < 3) {
            return text;
        }
        ArrayList<String> result = listMapChunksOfN(5, 3, tokens, al -> {
            // accepted patterns:
            // [glyph1, ws1, glyph2, ...] -> [glyph1, ws1, kern(glyph1, glyph2), ' ', glyph2, ...]
            // [glyph1, ws1, double, ws2, glyph2, ...] -> [glyph1, ws1, kern(glyph1, glyph2), ws2, glyph2, ...]
            // [space:glyph, ...] -> [space:glyph, ...] side effect: set kernSpaceAs to glyp
            if (isGlyphForKerning(al.get(0)) && isGlyphForKerning(al.get(2))) {
                al.add(2, formatKern(kernPair(al.get(0), al.get(2), kernSpaceAs[0])));
                al.add(3, " ");
            } else if (al.size() == 5
                    && isGlyphForKerning(al.get(0))
                    && PATTERN_POSITION_SIMPLE_X.matcher(al.get(2)).matches()
                    && isGlyphForKerning(al.get(4))) {
                al.set(2, formatKern(kernPair(al.get(0), al.get(4), kernSpaceAs[0])));
            } else {
                Matcher mKernSpaceAs = PATTERN_KERN_SPACE_AS.matcher(al.get(0));
                if (mKernSpaceAs.matches()) {
                    kernSpaceAs[0] = mKernSpaceAs.group(1);
                }
            }
        });
        return String.join("", result);
    }

    private boolean isGlyphForKerning(String token) {
        return glyphMapping.containsKey(token) || token.equals(GLYPH_SPACE);
    }

    private String textToTokens(String before, String text, String after) {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < text.length(); i++) {
            char c = text.charAt(i);
            if (charToGlyphName.containsKey(c + "")) {
                sb.append(charToGlyphName.get(c + "")).append(' ');
            } else if (Character.isLowerCase(c)) {
                sb.append(c).append(' ');
            } else if (Character.isUpperCase(c)) {
                sb.append(c).append(' ');
            } else if (Character.isDigit(c)) {
                sb.append(c).append('-').append(' ');
            } else if (c == '\n') {
                sb.append("\n0.0ax -0.8y\n");
                while (i + 1 < text.length() && text.charAt(i + 1) == ' ') {
                    i++;
                }
            } else if (c == ' ') {
                sb.append(" _ ");
                while (i + 1 < text.length() && text.charAt(i + 1) == ' ') {
                    i++;
                }
            } else if (c == '_') {
                sb.append(c);
                while (i + 1 < text.length() && text.charAt(i + 1) != ' ') {
                    sb.append(text.charAt(i + 1));
                    i++;
                }
            } else {
                sb.append(c).append(' ');
            }
        }
        String result = sb.toString();
        if (result.endsWith(" ")) {
            return result.substring(0, result.length() - 1);
        }
        return result;
    }

    public interface ModifySelection {
        String modify(String before, String selection, String after);
    }

    private void modifySelection(ModifySelection modify, boolean growSelectionToWholeTokens) {
        int selectionStart = textField.getSelectionStart();
        int selectionEnd = textField.getSelectionEnd();
        String text = textField.getText();
        if (growSelectionToWholeTokens) {
            selectionStart = growSelectionStart(selectionStart, text);
            selectionEnd = growSelectionEnd(selectionEnd, text);
        }
        String textBefore = text.substring(0, selectionStart);
        String textAfter = text.substring(selectionEnd);
        String textSelection = text.substring(selectionStart, selectionEnd);
        String modified = modify.modify(textBefore, textSelection, textAfter);
        textField.setText(textBefore + modified + textAfter);
        textField.setSelectionStart(selectionStart);
        textField.setSelectionEnd(selectionStart + modified.length());
        textField.requestFocus();
        reloadOrRefresh(false);
        repaint();
    }

    private int growSelectionEnd(int selectionEnd, String text) {
        while (selectionEnd < text.length() && selectionEnd > 0) {
            if (PATTERN_NONWS_NONWS.matcher(text.substring(selectionEnd - 1, selectionEnd + 1)).matches()) {
                selectionEnd++;
            } else {
                break;
            }
        }
        return selectionEnd;
    }

    private int growSelectionStart(int selectionStart, String text) {
        while (selectionStart > 0 && selectionStart < text.length()) {
            if (PATTERN_NONWS_NONWS.matcher(text.substring(selectionStart - 1, selectionStart + 1)).matches()) {
                selectionStart--;
            } else {
                break;
            }
        }
        return selectionStart;
    }

    private void open() {
        String directory = stateManager.getDirectory();
        FixedDirFileSystemView fsv = new FixedDirFileSystemView(new File(directory));
        JFileChooser jFileChooser = new JFileChooser(new File(directory), fsv);
        jFileChooser.setFileFilter(new FileFilter() {
            @Override
            public boolean accept(File f) {
                return f.isDirectory() || f.getName().endsWith(FILE_EXT_COMPOSITION);
            }

            @Override
            public String getDescription() {
                return FILE_EXT_COMPOSITION_DESCRIPTION;
            }
        });
        int result = jFileChooser.showDialog(this, "Load");
        if (result == JFileChooser.APPROVE_OPTION) {
            File selectedFile = jFileChooser.getSelectedFile();
            String fileNameNoExt = selectedFile.getName().replaceAll(Pattern.quote(FILE_EXT_COMPOSITION) + "\\z", "");
            nameTextField.setText(fileNameNoExt);
            load();
        }
    }

    private void load() {
        File file = new File(stateManager.getDirectory(), nameTextField.getText() + FILE_EXT_COMPOSITION);
        if (file.exists() && file.isFile()) {
            try {
                String text = new String(Files.readAllBytes(file.toPath()), StandardCharsets.UTF_8);
                textField.setText(text);
                reloadOrRefresh(true);
                repaint();
            } catch (IOException e) {
                JOptionPane.showMessageDialog(this, "Error reading file: " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }
        } else {
            JOptionPane.showMessageDialog(this, "File does not exists.", "Error", JOptionPane.ERROR_MESSAGE);
        }
    }

    private void save() {
        try {
            File fileComposition = new File(stateManager.getDirectory(), nameTextField.getText() + FILE_EXT_COMPOSITION);
            Files.write(fileComposition.toPath(), textField.getText().getBytes(StandardCharsets.UTF_8));
            try {
                File fileGlyphLocations = new File(stateManager.getDirectory(), nameTextField.getText() + FILE_EXT_GLYPH_LOCATIONS);
                File fileNumber = new File(stateManager.getDirectory(), nameTextField.getText() + FILE_EXT_NUMBER);

                saveGlyphLocationsAndNumber(new File(stateManager.getDirectory()), fileGlyphLocations, fileNumber);
            } catch (IOException | RuntimeException e) {
                JOptionPane.showMessageDialog(this, "Error saving glyph location / number file: " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }
        } catch (IOException | RuntimeException e) {
            JOptionPane.showMessageDialog(this, "Error saving file: " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        }
    }

    private void saveGlyphLocationsAndNumber(File glyphDir, File fileGlyphLocations, File fileNumber) throws IOException {
        Set<String> usedGlypgs = new TreeSet<>(Comparator.naturalOrder());
        usedGlypgs.addAll(Arrays.asList("0-", "1-", "2-", "3-", "4-", "5-", "6-", "7-", "8-", "9-"));
        for (GlyphLocation glyphLocation : glyphLocations) {
            usedGlypgs.add(glyphLocation.name);
        }
        System.out.println("used glyphs length: " + usedGlypgs.size());
        ArrayList<String> glyphNamesOrdered = new ArrayList<>(usedGlypgs);
        System.out.println("used glyphs: " + glyphNamesOrdered);

        Map<String, Integer> nameToIndex = new HashMap<>();
        for (int i = 0; i < glyphNamesOrdered.size(); i++) {
            String name = glyphNamesOrdered.get(i);
            nameToIndex.put(name, i);
        }

        NumberCompositor.TypesetGlyph[] typesetGlyphs = new NumberCompositor.TypesetGlyph[glyphLocations.size()];
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < glyphLocations.size(); i++) {
            GlyphLocation glyphLocation = glyphLocations.get(i);
            typesetGlyphs[i] = NumberCompositor.ts(nameToIndex.get(glyphLocation.name), glyphLocation.x, glyphLocation.y, glyphLocation.scaleDigit);
            sb.append(typesetGlyphs[i]);
            sb.append("\n");
        }
        String[] typesetGlyphNames = new String[glyphNamesOrdered.size()];
        for (int i = 0; i < typesetGlyphNames.length; i++) {
            typesetGlyphNames[i] = glyphMapping.get(glyphNamesOrdered.get(i)).glyphName;

        }

        String glyphLocationsStr = sb.toString();
        Files.write(fileGlyphLocations.toPath(), glyphLocationsStr.getBytes(StandardCharsets.UTF_8));

        String composedNumber = NumberCompositor.composeNumber(glyphDir.toString(), typesetGlyphNames, typesetGlyphs);
        Files.write(fileNumber.toPath(), composedNumber.getBytes(StandardCharsets.UTF_8));
    }

    private void exportView() {
        double exportScale = 5;
        BufferedImage b = new BufferedImage((int) (renderPanel.getWidth() * exportScale), (int) (renderPanel.getHeight() * exportScale), BufferedImage.TYPE_BYTE_GRAY);
        paintGlyphs(b.createGraphics(), b.getWidth(), b.getHeight(), true, glyphScale * exportScale);
        try {
            ImageIO.write(b, "png", new File(stateManager.getDirectory(),"exportView_" + System.currentTimeMillis() + ".png"));
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void paintCursor(Graphics g, int width, int height, double glyphScale) {
        int selectionStart = textField.getSelectionStart();
        int selectionEnd = textField.getSelectionEnd();
        if (selectionStart == selectionEnd) {
            paintCursorThin(g, height, glyphScale, selectionStart);
        } else {
            int glyphCountLast = -1;
            for (int selectionPos = selectionStart; selectionPos < selectionEnd; selectionPos++) {
                if (sourceLocationInfo != null && sourceLocationInfo.getLength() > selectionPos) {
                    int glyphCount = sourceLocationInfo.getGlyphCount(selectionPos);
                    if (glyphCount != glyphCountLast) { // todo
                        paintCursorBlock(g, height, glyphScale, selectionPos);
                        glyphCountLast = glyphCount;
                    }
                }
            }
        }
    }

    private void paintCursorThin(Graphics g, int height, double glyphScale, int textIndex) {
        if (sourceLocationInfo != null && sourceLocationInfo.getLength() > textIndex) {
            double selX = sourceLocationInfo.getX(textIndex);
            double selY = sourceLocationInfo.getY(textIndex);
            double selScale = sourceLocationInfo.getScale(textIndex);
            double selAX = sourceLocationInfo.getAnchorX(textIndex);
            double posSelX = (selX - glyphViewX + selAX - 0.05) * glyphPixels * glyphScale;
            double posSelY = (selY - glyphViewY) * glyphPixels * glyphScale;
            int targetHeight = (int) (glyphPixels * selScale * glyphScale);
            int targetWidth = (int) (glyphPixels * selScale * glyphScale * 0.07);
            int posSelXint = (int) (posSelX);
            int posSelYint = height - targetHeight - (int) (posSelY);
            g.setXORMode(Color.yellow);
            g.fillRect(posSelXint, posSelYint, targetWidth, targetHeight);
        }
    }

    private void paintCursorBlock(Graphics g, int height, double glyphScale, int textIndex) {
        if (sourceLocationInfo != null && sourceLocationInfo.getLength() > textIndex) {
            double selX = sourceLocationInfo.getX(textIndex);
            double selY = sourceLocationInfo.getY(textIndex);
            double selScale = sourceLocationInfo.getScale(textIndex);
            String glyphName = sourceLocationInfo.getGlyphName(textIndex);
            GlyphBitmap glyphBitmap = glyphMapping.get(glyphName);
            if (glyphBitmap != null) {
                double posSelXMin = ((selX - glyphViewX) * glyphPixels + glyphBitmap.bbMinX * selScale) * glyphScale;
                double posSelYMin = ((selY - glyphViewY) * glyphPixels + glyphBitmap.bbMinY * selScale) * glyphScale;
                double posSelWidth = ((glyphBitmap.bbMaxX - glyphBitmap.bbMinX) * selScale) * glyphScale;
                double posSelHeight = ((glyphBitmap.bbMaxY - glyphBitmap.bbMinY) * selScale) * glyphScale;
                int posSelXMinint = (int) (posSelXMin);
                int posSelYMinint = height - (int) posSelYMin - (int) posSelHeight;
                g.setXORMode(Color.yellow);
                g.fillRect(posSelXMinint - 1, posSelYMinint - 1, (int) posSelWidth + 2, (int) posSelHeight + 2);
            }
        }
    }

    private void paintGlyphs(Graphics g, int width, int height, boolean highQuality, double glyphScale) {
        Rectangle clipBounds = g.getClipBounds();
        if (clipBounds == null) {
            clipBounds = new Rectangle(0, 0, width, height);
        }
        Graphics2D g2 = (Graphics2D) g;
        g2.setBackground(Color.white);
        g2.clearRect(clipBounds.x, clipBounds.y, clipBounds.width, clipBounds.height);
        for (int i = 0; i < glyphLocations.size(); i++) {
            GlyphLocation glyphLocation = glyphLocations.get(i);
            System.out.println("step " + (i + 1) + " of " + glyphLocations.size());
            paintGlyph(g, height, highQuality, glyphScale, g2, glyphLocation);
        }
    }

    private void paintGlyph(Graphics g, int height, boolean highQuality, double glyphScale, Graphics2D g2, GlyphLocation glyphLocation) {
        int scaleDigit = glyphLocation.scaleDigit;
        double currentGlyphScale = scaleDigitToScale(scaleDigit);
        g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, currentGlyphScale < 1 ? RenderingHints.VALUE_INTERPOLATION_BILINEAR : RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
        GlyphBitmap glyphBitmap = glyphMapping.get(glyphLocation.name);
        int targetHeight = (int) (glyphPixels * currentGlyphScale * glyphScale);
        int targetWidth = (int) (glyphPixels * currentGlyphScale * glyphScale);
        BufferedImage bi = glyphBitmap.bi;
        Image scaledInstance = highQuality ? scaledInstanceCache(() -> bi.getScaledInstance(targetWidth, targetHeight, Image.SCALE_AREA_AVERAGING), bi, targetWidth, targetHeight) : bi;

        double posX = (glyphLocation.x - glyphViewX) * glyphPixels * glyphScale;
        double posY = (glyphLocation.y - glyphViewY) * glyphPixels * glyphScale;
        int posXint = (int) (posX);
        int posYint = height - targetHeight - (int) (posY);
        g.drawImage(scaledInstance,
                posXint,
                posYint,
                targetWidth,
                targetHeight,
                null);
    }

    private WeakHashMap<Image, Map<Integer, Map<Integer, Image>>> scaledInstanceCacheMap = new WeakHashMap<>();

    private Image scaledInstanceCache(Supplier<Image> imageSupplier, Image sourceImage, int targetWidth, int targetHeight) {
        Map<Integer, Map<Integer, Image>> integerMapMap = scaledInstanceCacheMap.get(sourceImage);
        if (integerMapMap != null) {
            Map<Integer, Image> integerImageMap = integerMapMap.get(targetWidth);
            if (integerImageMap != null) {
                Image image = integerImageMap.get(targetHeight);
                if (image != null) {
                    return image;
                } else {
                    image = imageSupplier.get();
                    integerImageMap.put(targetHeight, image);
                    return image;
                }
            } else {
                Image image = imageSupplier.get();
                integerImageMap = new HashMap<>();
                integerImageMap.put(targetHeight, image);
                integerMapMap.put(targetWidth, integerImageMap);
                return image;
            }
        } else {
            Image image = imageSupplier.get();
            Map<Integer, Image> integerImageMap = new HashMap<>();
            integerImageMap.put(targetHeight, image);
            integerMapMap = new HashMap<>();
            integerMapMap.put(targetWidth, integerImageMap);
            scaledInstanceCacheMap.put(sourceImage, integerMapMap);
            return image;
        }
    }

    public static double scaleDigitToScale(int scaleDigit) {
        return Math.pow(1.25d, scaleDigit - 5);
    }

    static class GlyphBitmap {
        public static final String FILE_EXT_COMPOSITOR_CACHE = ".compositor-cache.png";
        final String glyphName;
        final BufferedImage bi;
        final int bbMinX, bbMaxX, bbMinY, bbMaxY;

        public GlyphBitmap(String glyphName, BufferedImage bi, int bbMinX, int bbMaxX, int bbMinY, int bbMaxY) {
            this.glyphName = glyphName;
            this.bi = bi;
            this.bbMinX = bbMinX;
            this.bbMaxX = bbMaxX;
            this.bbMinY = bbMinY;
            this.bbMaxY = bbMaxY;
        }

        public void saveCache(String directory, String glyphName) throws IOException {
            ImageIO.write(bi, "png", new File(directory, glyphName + FILE_EXT_COMPOSITOR_CACHE));
        }

        public static GlyphBitmap loadCache(String directory, String glyphName) throws IOException {
            File fileCache = new File(directory, glyphName + FILE_EXT_COMPOSITOR_CACHE);
            if (!fileCache.exists()) {
                return null;
            }
            File fileGlyph = new File(directory, glyphName + StateManager.FILE_EXT_GLYPH);
            if (!fileGlyph.exists()) {
                return null;
            }
            if (fileCache.lastModified() > fileGlyph.lastModified()) {
                System.out.println("using glyph bitmap from cache: " + glyphName);
                return load(fileCache, glyphName);
            }
            System.out.println("obsolete glyph bitmap in cache: " + glyphName);
            return null;
        }

        private static GlyphBitmap load(File fileCache, String glyphName) throws IOException {
            BufferedImage bi = ImageIO.read(fileCache);
            int h = bi.getHeight();
            int w = bi.getWidth();
            int bbMinX = w, bbMaxX = 0, bbMinY = h, bbMaxY = 0;
            WritableRaster r = bi.getRaster();
            int[] iArray = new int[4];
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++) {
                    r.getPixel(x, y, iArray);
                    if (iArray[3] > 0) {
                        bbMinX = Math.min(bbMinX, x);
                        bbMaxX = Math.max(bbMaxX, x);
                        bbMinY = Math.min(bbMinY, y);
                        bbMaxY = Math.max(bbMaxY, y);
                    }
                }
            }
            return new GlyphBitmap(glyphName, bi, bbMinX, bbMaxX, h - bbMaxY, h - bbMinY);
        }
    }

    static class GlyphLocation {
        String name;
        double x, y;
        int scaleDigit;

        public GlyphLocation(String name, double x, double y, int scaleDigit) {
            this.name = name;
            this.x = x;
            this.y = y;
            this.scaleDigit = scaleDigit;
        }

        @Override
        public String toString() {
            return "GlyphLocation{" +
                    "name='" + name + '\'' +
                    ", x=" + x +
                    ", y=" + y +
                    ", scaleDigit=" + scaleDigit +
                    '}';
        }
    }

    class SourceLocationInfo {
        private int count;
        private double[] sourceLocationsX;
        private double[] sourceLocationsY;
        private double[] sourceLocationsScale;
        private double[] sourceLocationsAnchorX;
        private double[] sourceLocationsAnchorY;
        private int[] sourceLocationsGlyphCounter;
        private String[] sourceLocationsGlyphName;
        private int position;

        SourceLocationInfo(int length) {
            sourceLocationsX = new double[length + 1];
            sourceLocationsY = new double[length + 1];
            sourceLocationsScale = new double[length + 1];
            sourceLocationsAnchorX = new double[length + 1];
            sourceLocationsAnchorY = new double[length + 1];
            sourceLocationsGlyphCounter = new int[length + 1];
            sourceLocationsGlyphName = new String[length + 1];
            position = 0;
            count = 0;
        }

        public void update(String token, double x, double y, int scaleDigit, boolean glyph) {
            int length = token.length();
            double scale = scaleDigitToScale(scaleDigit);
            x = x - glyphAnchorX * scale;
            y = y - glyphAnchorY * scale;
            if (glyph) {
                count++;
            }
            for (int i = 0; i < length; i++) {
                sourceLocationsX[position] = x;
                sourceLocationsY[position] = y;
                sourceLocationsScale[position] = scale;
                sourceLocationsAnchorX[position] = glyphAnchorX;
                sourceLocationsAnchorY[position] = glyphAnchorY;
                sourceLocationsGlyphCounter[position] = count;
                sourceLocationsGlyphName[position] = token;
                position++;
            }
        }

        public int getLength() {
            return sourceLocationsX.length;
        }

        public double getX(int i) {
            return sourceLocationsX[i];
        }

        public double getY(int i) {
            return sourceLocationsY[i];
        }

        public double getScale(int i) {
            return sourceLocationsScale[i];
        }

        public double getAnchorX(int i) {
            return sourceLocationsAnchorX[i];
        }

        public double getAnchorY(int i) {
            return sourceLocationsAnchorY[i];
        }

        public String getGlyphName(int i) {
            return sourceLocationsGlyphName[i];
        }

        public int getGlyphCount(int i) {
            return sourceLocationsGlyphCounter[i];
        }
    }

    /**
     * example input: a=a11 b=b2 c=c1 a 0.3 b 0.4y c 0.1ax a 0.2 b
     */
    private void refreshAction() {
        reloadOrRefresh(true);
        repaint();
    }

    private void reloadOrRefresh(boolean reload) {
        renderedTextBitmap = null;
        if (reload) {
            glyphMapping.clear();
            kernPreProcessCache.clear();
            scaledInstanceCacheMap.clear();
        }
        glyphAnchorX = 0.2d;
        glyphAnchorY = 0.3d;
        System.out.println("refresh started reload=" + reload);
        String text = textField.getText();
        List<String> tokens = tokenize(text);
        scanAndLoadGlyphs(tokens);
        scanAndSetGlobals(tokens);
        glyphLocations.clear();
        sourceLocationInfo = new SourceLocationInfo(text.length());
        double ax = 0.0d, ay = 0.0d;
        int scaleDigit = 5;
        double scale = 1;
        for (int i = 0; i < tokens.size(); i++) {
            String expression = tokens.get(i);

            if (PATTERN_WS_PLUS.matcher(expression).matches()) {
                sourceLocationInfo.update(expression, ax, ay, scaleDigit, false);
                continue;
            }
            String[] charToFile = expression.split("=");
            if (charToFile.length != 2) {
                Matcher matcher = PATTERN_POSITION.matcher(expression);
                if (matcher.matches()) {
                    String numberStr = matcher.group(1);
                    double number = Double.parseDouble(numberStr);
                    String axis = matcher.group(4);
                    if (axis == null) {
                        axis = "x";
                    }
                    if (axis.equals("x")) {
                        ax += number * scale;
                    } else if (axis.equals("ax")) {
                        ax = number;
                    } else if (axis.equals("y")) {
                        ay += number * scale;
                    } else if (axis.equals("ay")) {
                        ay = number;
                    }
                    sourceLocationInfo.update(expression, ax, ay, scaleDigit, false);
                    continue;
                }
                Matcher matcherScaleDigit = PATTERN_SCALE_DIGIT.matcher(expression);
                if (matcherScaleDigit.matches()) {
                    String scaleDigitStr = matcherScaleDigit.group(1);
                    scaleDigit = Integer.parseInt(scaleDigitStr);
                    scale = scaleDigitToScale(scaleDigit);
                    sourceLocationInfo.update(expression, ax, ay, scaleDigit, false);
                    continue;
                }
                Matcher matcherGlyphAnchor = PATTERN_GLYPH_ANCHOR.matcher(expression);
                if (matcherGlyphAnchor.matches()) {
                    String glyphAnchorXStr = matcherGlyphAnchor.group(1);
                    String glyphAnchorYStr = matcherGlyphAnchor.group(2);
                    glyphAnchorX = Double.parseDouble(glyphAnchorXStr);
                    glyphAnchorY = Double.parseDouble(glyphAnchorYStr);
                    sourceLocationInfo.update(expression, ax, ay, scaleDigit, false);
                    continue;
                }
                // no special command, add glyph location
                if (glyphMapping.containsKey(expression)) {
                    glyphLocations.add(new GlyphLocation(expression, ax - glyphAnchorX * scale, ay - glyphAnchorY * scale, scaleDigit));
                    sourceLocationInfo.update(expression, ax, ay, scaleDigit, true);
                    continue;
                } else {
                    System.out.println("unknown expression: '" + expression + "'");
                }
            }
            sourceLocationInfo.update(expression, ax, ay, scaleDigit, false);
        }
        sourceLocationInfo.update(" ", ax, ay, scaleDigit, false);
        System.out.println("glyphLocations: " + glyphLocations);
        System.out.println("refresh done");
        if (reload) {
            System.out.println("loaded count=" + glyphMapping.size());
        }
    }

    public void scanTokens(List<String> tokens, Pattern pattern, Consumer<Matcher> consume) {
        for (int i = 0; i < tokens.size(); i += 2) {
            String token = tokens.get(i);
            Matcher matcher = pattern.matcher(token);
            if (matcher.matches()) {
                consume.accept(matcher);
            }
        }
    }

    private void scanAndSetGlobals(List<String> tokens) {
        scanTokens(tokens, PATTERN_SCALE, m -> {
            glyphScale = Double.parseDouble(m.group(1));
        });
        scanTokens(tokens, PATTERN_VIEW, m -> {
            glyphViewX = Double.parseDouble(m.group(1));
            glyphViewY = Double.parseDouble(m.group(2));
        });
    }


    private void scanAndLoadGlyphs(List<String> tokens) {
        // load glyphs
        scanTokens(tokens, PATTERN_GLYPH_MAPPING, m -> {
            String glyphReference = m.group(1);
            String glyphName = m.group(2);
            GlyphBitmap glyphBitmap = glyphMapping.get(glyphReference);
            if (glyphBitmap == null || !glyphBitmap.glyphName.equals(glyphName)) {
                try {
                    glyphMapping.put(glyphReference, loadGlyphBitmap(glyphName));
                    kernPreProcessCache.remove(glyphReference);
                } catch (IOException | RuntimeException e) {
                    System.out.println("could not load glyphName: " + glyphName + " due to " + e);
                    e.printStackTrace();
                }
            }
        });
    }

    /**
     * get string split by white space including delimiters
     *
     * @param text
     * @return ArrayList of length divisible by 2; element 0 is token, element 1 is whitespace-delimiter, element 1 is token ... (first and last token may be empty)
     */
    public static ArrayList<String> tokenize(String text) {
        return new ArrayList<>(Arrays.asList(PATTERN_TOKEN_DELIMITER.matcher(text).replaceAll("\u0001$1\u0001").split("\u0001")));
    }

    private static Map<String, String> charToGlyphName = new HashMap<>();

    static {
        charToGlyphName.put(".", "_dot");
        charToGlyphName.put(",", "_comm");
        charToGlyphName.put("+", "_plus");
        charToGlyphName.put("-", "_ndash");
        charToGlyphName.put("*", "_times");
        charToGlyphName.put("/", "_slash");
        charToGlyphName.put("=", "_equal");
        charToGlyphName.put("(", "_parl");
        charToGlyphName.put(")", "_parr");
        charToGlyphName.put("'", "_apos");
        charToGlyphName.put("<", "_lt");
        charToGlyphName.put(">", "_gt");
        charToGlyphName.put("á", "ac");
        charToGlyphName.put("é", "ec");
        charToGlyphName.put("í", "ic");
    }

    private double kernPair(String glyph1, String glyph2, String kernSpaceAs) {
        if (glyph1.equals(GLYPH_SPACE)) {
            glyph1 = kernSpaceAs;
        }
        if (glyph2.equals(GLYPH_SPACE)) {
            glyph2 = kernSpaceAs;
        }
        GlyphBitmap glyphBitmap1 = glyphMapping.get(glyph1);
        GlyphBitmap glyphBitmap2 = glyphMapping.get(glyph2);
        if (glyphBitmap1 == null || glyphBitmap2 == null) {
            return 0.5;
        }
        BufferedImage bi1 = glyphBitmap1.bi;
        BufferedImage bi2 = glyphBitmap2.bi;
        Kerning k = new BlurringKerning4(4000, 20, 0, 45, 0);
        Object kp1 = kernPreProcessCache.get(glyph1);
        if (kp1 == null) {
            kp1 = k.kernPreProcess(bi1, glyph1);
            kernPreProcessCache.put(glyph1, kp1);
        }
        Object kp2 = kernPreProcessCache.get(glyph2);
        if (kp2 == null) {
            kp2 = k.kernPreProcess(bi2, glyph2);
            kernPreProcessCache.put(glyph2, kp2);
        }
        int kernPixels = k.kernEval(kp1, kp2);
        double kern = ((double) kernPixels) / glyphPixels;
        return kern;
    }

    public static String formatKern(double kern) {
        return String.format(Locale.ENGLISH, "%.3f", kern);
    }


    private GlyphBitmap loadGlyphBitmap(String glyphName) throws IOException {
        String directory = this.stateManager.getDirectory();
        GlyphBitmap cached = GlyphBitmap.loadCache(directory, glyphName);
        if (cached != null) {
            return cached;
        }
        GlyphDesignerModel loadGlyphDesignerModel = new GlyphDesignerModel();
        Grid grid = new Grid();
        StateManager loadStateManager = new StateManager(loadGlyphDesignerModel, grid);
        loadStateManager.setDirectory(directory);
        loadStateManager.setGlyphName(glyphName);
        loadStateManager.loadAction();
        int glyphPixelsMul2 = 2 * glyphPixels;
        BufferedImage bi = new BufferedImage(glyphPixelsMul2, glyphPixelsMul2, BufferedImage.TYPE_3BYTE_BGR);
        Graphics2D g = (Graphics2D) bi.getGraphics();
        g.setBackground(new Color(255, 255, 255, 0));
        g.clearRect(0, 0, glyphPixelsMul2, glyphPixelsMul2);
        GlyphPainter.drawGlyph(new Dimension(glyphPixelsMul2, glyphPixelsMul2), bi, loadGlyphDesignerModel.glyph, grid, new ViewZoom(), new GlyphPainter.GlyphPainterCache(glyphPixelsMul2, glyphPixelsMul2), null, false, false);
        long t1 = System.nanoTime();
        BufferedImage bi2 = new BufferedImage(glyphPixels, glyphPixels, BufferedImage.TYPE_4BYTE_ABGR);
        WritableRaster r = bi.getRaster();
        WritableRaster r2 = bi2.getRaster();
        int w = r2.getWidth();
        int h = r2.getHeight();
        int[] ca = new int[]{0, 0, 0, 255};
        int[] camap = new int[]{0, 63, 127, 191, 255};
        int[] ar3 = new int[3];
        int bbMinX = w, bbMaxX = 0, bbMinY = h, bbMaxY = 0;
        for (int y = 0; y < h; y++) {
            for (int x = 0; x < w; x++) {
                int c = 0;
                r.getPixel(x * 2, y * 2, ar3);
                c += ar3[0] == 255 ? 0 : 1;
                r.getPixel(x * 2 + 1, y * 2, ar3);
                c += ar3[0] == 255 ? 0 : 1;
                r.getPixel(x * 2, y * 2 + 1, ar3);
                c += ar3[0] == 255 ? 0 : 1;
                r.getPixel(x * 2 + 1, y * 2 + 1, ar3);
                c += ar3[0] == 255 ? 0 : 1;
                ca[3] = camap[c];
                r2.setPixel(x, y, ca);
                if (c > 0) {
                    bbMinX = Math.min(bbMinX, x);
                    bbMaxX = Math.max(bbMaxX, x);
                    bbMinY = Math.min(bbMinY, y);
                    bbMaxY = Math.max(bbMaxY, y);
                }
            }
        }
        long t2 = System.nanoTime();
        System.out.println("loadGlyphBitmap took " + ((t2 - t1) / 1e9) + "s for " + loadGlyphDesignerModel.glyph.curves.size() + " curves.");
        GlyphBitmap glyphBitmap = new GlyphBitmap(glyphName, bi2, bbMinX, bbMaxX, h - bbMaxY, h - bbMinY);
        glyphBitmap.saveCache(directory, glyphName);
        return glyphBitmap;
    }

}
