/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;

/**
 */
public class Curve {
    public Point p1;
    public Point p2;
    public Point p3;
    public Point p4;
    public boolean selected;

    public Curve() {
    }

    public Curve(Point p1, Point p2, Point p3, Point p4, boolean selected) {
        this.p1 = p1;
        this.p2 = p2;
        this.p3 = p3;
        this.p4 = p4;
        this.selected = selected;
    }

    public List<Point> allPoints() {
        return Arrays.asList(p1, p2, p4, p3);
    }

    public List<Point> allEndPoints() {
        return Arrays.asList(p1, p4);
    }

    public List<Point> allCurveControlPoints() {
        return Arrays.asList(p2, p3);
    }

    public List<Point> allAssociatedPoints(Point p, PointRelation pr) {
        ArrayList<Point> ap = new ArrayList<>();
        if (p == p1) {
            if (pr.apply(p1, p2)) {
                ap.add(p2);
            }
        }
        if (p == p4) {
            if (pr.apply(p4, p3)) {
                ap.add(p3);
            }
        }
        return ap;
    }

    public String describePoint(Point p) {
        if (p == p1) {
            return "P1";
        } else if (p == p2) {
            return "P2";
        } else if (p == p3) {
            return "P3";
        } else if (p == p4) {
            return "P4";
        } else {
            return null;
        }
    }

    @Override
    public String toString() {
        return "Curve{" +
                ", p1=" + p1 +
                ", p2=" + p2 +
                ", p3=" + p3 +
                ", p4=" + p4 +
                ", selected=" + selected +
                '}';
    }

    public Curve duplicate() {
        Curve curve = new Curve();
        curve.p1 = p1.duplicate();
        curve.p2 = p2.duplicate();
        curve.p3 = p3.duplicate();
        curve.p4 = p4.duplicate();
        curve.selected = selected;
        return curve;
    }

    public void deselect() {
        selected = false;
        for (Point p : allPoints()) {
            p.selected = false;
        }
    }

    public void select() {
        selected = true;
    }

    public void moveRel(double dx, double dy) {
        allPoints().forEach(p -> p.moveRel(dx, dy));
    }

    public void rotateCCW(MousePos pivot) {
        allPoints().forEach(p -> p.rotateCCW(pivot));
    }

    public void rotateCW(MousePos pivot) {
        allPoints().forEach(p -> p.rotateCW(pivot));
    }

    public void flipLeftRight(MousePos pivot) {
        allPoints().forEach(p -> p.flipLeftRight(pivot));
    }

    public void flipTopBottom(MousePos pivot) {
        allPoints().forEach(p -> p.flipTopBottom(pivot));
    }


    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Curve curve = (Curve) o;
        return Objects.equals(p1, curve.p1) &&
                Objects.equals(p2, curve.p2) &&
                Objects.equals(p3, curve.p3) &&
                Objects.equals(p4, curve.p4);
    }

    @Override
    public int hashCode() {
        return Objects.hash(p1, p2, p3, p4);
    }

    public boolean contains(Point p) {
        return p == p1 || p == p2 || p == p3 || p == p4;
    }
}
