/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import java.util.Arrays;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Objects;

/**
 *
 */
public class EventQueue {
    @FunctionalInterface
    public interface EventConsumer {
        void consumeEvent(Object o);
    }

    private LinkedHashSet<EventConsumer> eventConsumers = new LinkedHashSet<>();
    private Deque<Object> mergeList;

    public void postEvent() {
        postEvent(null);
    }

    public synchronized void postEvent(Object o) {
        if (mergeList != null) {
            mergeList.add(o);
        } else {
            callConsumers(o);
        }
    }

    private void callConsumers(Object o) {
        for (EventConsumer eventConsumer : new LinkedHashSet<>(eventConsumers)) {
            eventConsumer.consumeEvent(o);
        }
    }

    public synchronized void addEventConsumer(EventConsumer eventConsumer) {
        eventConsumers.add(eventConsumer);
    }

    public synchronized void removeEventConsumer(EventConsumer eventConsumer) {
        eventConsumers.remove(eventConsumer);
    }

    public synchronized void removeAllEventConsumers() {
        eventConsumers.clear();
    }

    public synchronized void mergeEventsDuring(Runnable runnable) {
        if (mergeList != null) {
            mergeList = new LinkedList<>();
            runnable.run();
            while (!mergeList.isEmpty()) {
                Object first = mergeList.pop();
                if (mergeList.isEmpty()) {
                    callConsumers(first);
                } else {
                    if (!Objects.equals(first, mergeList.poll())) {
                        callConsumers(first);
                    }
                }
            }
            mergeList = null;
        } else {
            runnable.run();
        }
    }

    public static void multiMergeEventsDuring(Runnable runnable, EventQueue... meqs) {
        multiMergeEventsDuring(runnable, Arrays.asList(meqs).iterator());
    }

    private static void multiMergeEventsDuring(Runnable runnable, Iterator<EventQueue> eqs) {
        if (eqs.hasNext()) {
            eqs.next().mergeEventsDuring(() -> multiMergeEventsDuring(runnable, eqs));
        } else {
            runnable.run();
        }
    }

}
