/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileSystemView;
import java.awt.*;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;
import java.util.function.*;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 *
 */
public class GlyphDesigner extends JFrame {

    private final JPanel drawPanel;
    private final Grid grid;
    private final ViewZoom viewZoom;
    private final HistoryLister historyLister;
    private final Compositor compositor;
    private GlyphDesignerModel glyphDesignerModel;
    private JTextField detailXTextField;
    private JTextField detailYTextField;
    private JLabel detailInfoLabel;
    private GlyphPainter glyphPainter;
    private JTextField[] xGridDivisionsTextField;
    private JTextField[] yGridDivisionsTextField;
    private ButtonGroup activeGridGroup;
    private JRadioButton[] activeGridRadioButton;
    private JList<String> curveList;
    private boolean userUpdatesPoints = true;
    private JTextField x1ZoomTextField, x2ZoomTextField, y1ZoomTextField, y2ZoomTextField;
    private StateManager stateManager;
    private ArrayList<Preview> siblings;
    private JTextField glyphNameField;


    public GlyphDesigner() throws HeadlessException {
        this.setTitle("Glyph Designer (outline wn)");
        this.setBounds(200, 40, 1200, 900);

        this.grid = new Grid();
        viewZoom = new ViewZoom();
        glyphPainter = new GlyphPainter(viewZoom);
        grid.getUpdateEventQueue().addEventConsumer(o -> repaintDrawPanel());
        viewZoom.getUpdateEventQueue().addEventConsumer(o -> repaintDrawPanelInvalidateCache());

        this.glyphDesignerModel = new GlyphDesignerModel();
        stateManager = new StateManager(glyphDesignerModel, grid);

        drawPanel = new JPanel() {
            @Override
            protected void paintComponent(Graphics g) {
                super.paintComponent(g);
                Dimension size = this.getSize();
                glyphPainter.paintOnGraphics(size, g, glyphDesignerModel.glyph, grid, (BufferedImage bi) -> {
                    for (Preview preview : siblings) {
                        preview.updateViewForShiftByMasterView(viewZoom);
                        int w = (int) Math.round(size.getWidth());
                        int h = (int) Math.round(size.getHeight());
                        GlyphPainter.drawGlyph(drawPanel.getSize(), bi, preview.previewGlyphDesignerModel.glyph, preview.previewGrid, preview.viewZoom, preview.previewGlyphPainter.getCache(w, h), null, true, false);
                    }
                });
                MousePos rectangleSelection = glyphDesignerModel.getRectangleSelection();
                if (rectangleSelection != null) {
                    MousePos rectangleSelectionEnd = glyphDesignerModel.getRectangleSelectionEnd();
                    glyphPainter.paintSelectionRectangle(size, g, rectangleSelection.x, rectangleSelection.y, rectangleSelectionEnd.x, rectangleSelectionEnd.y);
                }
            }

        };
        drawPanel.setFocusable(true);
        drawPanel.setRequestFocusEnabled(true);
        MouseAdapter mouseAdapter = new MouseAdapter() {
            public int cycler;

            @Override
            public void mouseClicked(MouseEvent e) {
                e.getComponent().requestFocusInWindow();
                MousePos m = new CoordinateMapping(drawPanel.getSize(), viewZoom).mouseEventToMousePos(e);
                if (!glyphDesignerModel.handleClick(grid, m)) {
                    // find using selectionHint
                    ArrayList<Integer> hint = glyphPainter.getSelectionHintFor(e.getX(), e.getY());
                    cycler++;
                    if (!hint.isEmpty()) {
                        int idx = hint.get(cycler % hint.size());
                        if (idx == glyphDesignerModel.getSelectedIndex()) {
                            cycler++;
                            idx = hint.get(cycler % hint.size());
                        }
                        if (m.mod1) {
                            if (curveList.isSelectedIndex(idx)) {
                                glyphDesignerModel.removeCurveFromSelection(idx);
                            } else {
                                glyphDesignerModel.addCurveToSelection(idx);
                            }
                        } else {
                            glyphDesignerModel.selectCurveMulti(idx);
                        }
                    } else {
                        if (m == null || !m.mod1) {
                            glyphDesignerModel.selectCurveNone();
                        }
                    }
                }
            }

            @Override
            public void mouseMoved(MouseEvent e) {
                glyphDesignerModel.handlePosition(grid, new CoordinateMapping(drawPanel.getSize(), viewZoom).mouseEventToMousePos(e));
            }

            @Override
            public void mouseDragged(MouseEvent e) {
                glyphDesignerModel.handleDrag(grid, new CoordinateMapping(drawPanel.getSize(), viewZoom).mouseEventToMousePos(e));
            }

            @Override
            public void mousePressed(MouseEvent e) {
                stateManager.mouseStart();
                glyphDesignerModel.handlePress(grid, new CoordinateMapping(drawPanel.getSize(), viewZoom).mouseEventToMousePos(e));
            }

            @Override
            public void mouseReleased(MouseEvent e) {
                glyphDesignerModel.handleRelease(grid, new CoordinateMapping(drawPanel.getSize(), viewZoom).mouseEventToMousePos(e));
                stateManager.mouseEnd();
            }
        };
        drawPanel.addMouseListener(mouseAdapter);
        drawPanel.addMouseMotionListener(mouseAdapter);
        drawPanel.addFocusListener(new FocusListener() {
            @Override
            public void focusGained(FocusEvent e) {
            }

            @Override
            public void focusLost(FocusEvent e) {
                glyphDesignerModel.handleCancel();
            }
        });

        JScrollPane jScrollPane = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);

        JPanel rightColumn = new JPanel();
        rightColumn.setLayout(new BoxLayout(rightColumn, BoxLayout.Y_AXIS));

        jScrollPane.setViewportView(rightColumn);

        JPanel detailFilePanel = new JPanel();
        detailFilePanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "File"), BorderFactory.createEmptyBorder(6, 10, 8, 10)));
        detailFileComponents(detailFilePanel);
        rightColumn.add(detailFilePanel);

        JPanel detailGridPanel = new JPanel();
        detailGridPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Grid/Guides setup"), BorderFactory.createEmptyBorder(6, 10, 8, 10)));
        detailGridComponents(detailGridPanel);
        rightColumn.add(detailGridPanel);

        JPanel detailZoomPanel = new JPanel();
        detailZoomPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Viewport"), BorderFactory.createEmptyBorder(6, 10, 8, 10)));
        detailZoomComponents(detailZoomPanel);
        rightColumn.add(detailZoomPanel);

        JPanel detailPointPanel = new JPanel();
        detailPointPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Point detail"), BorderFactory.createEmptyBorder(6, 10, 8, 10)));
        detailPointComponents(detailPointPanel);
        this.glyphDesignerModel.getChangesToCurves().addEventConsumer((obj) -> updateDetailPointUI());
        this.glyphDesignerModel.getChangesToLayout().addEventConsumer((obj) -> updateDetailPointUI());
        rightColumn.add(detailPointPanel);
        updateDetailPointUI();

        JPanel selectedPointsPanel = new JPanel();
        selectedPointsPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Selected points actions"), BorderFactory.createEmptyBorder(6, 10, 8, 10)));
        selectedPointsComponents(selectedPointsPanel);
        rightColumn.add(selectedPointsPanel);

        JPanel selectedCurvesPanel = new JPanel();
        selectedCurvesPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Selected curve actions"), BorderFactory.createEmptyBorder(6, 10, 8, 10)));
        selectedCurvesComponents(selectedCurvesPanel);
        rightColumn.add(selectedCurvesPanel);

        JPanel curveListPanel = new JPanel();
        curveListPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Curve list"), BorderFactory.createEmptyBorder(6, 10, 8, 10)));
        curveListComponents(curveListPanel);
        rightColumn.add(curveListPanel);

        this.setLayout(new BorderLayout(4, 4));
        this.add(drawPanel, BorderLayout.CENTER);
        this.add(jScrollPane, BorderLayout.EAST);
        this.setVisible(true);
        this.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        this.historyLister = new HistoryLister(stateManager);
        this.compositor = new Compositor(stateManager);
    }

    private void detailFileComponents(JPanel detailFilePanel) {
        detailFilePanel.setLayout(new BoxLayout(detailFilePanel, BoxLayout.Y_AXIS));

        JPanel detailFileTopPanel = new JPanel();
        detailFileTopPanel.setLayout(new GridBagLayout());
        GridBagConstraints c = gridBagInit();
        gridBagAdder(detailFileTopPanel, new JLabel("Directory: "), c);
        JTextField dirTextField = new JTextField(".", 20);
        textFieldOnChange(dirTextField, text -> stateManager.setDirectory(text));
        c.fill = GridBagConstraints.HORIZONTAL;
        gridBagAdder(detailFileTopPanel, dirTextField, c);
        c.fill = 0;
        JButton dirChooseButton = new JButton("Choose");
        dirChooseButton.addActionListener(e -> {
            JFileChooser jFileChooser = new JFileChooser(new File(stateManager.getDirectory()));
            jFileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
            int result = jFileChooser.showDialog(detailFilePanel, "Select");
            if (result == JFileChooser.APPROVE_OPTION) {
                File selectedFile = jFileChooser.getSelectedFile();
                for (Preview sibling : siblings) {
                    sibling.previewStateManager.setDirectory(selectedFile.toString());
                }
                dirTextField.setText(selectedFile.toString());
                dirTextField.postActionEvent();
            }
        });
        gridBagAdder(detailFileTopPanel, dirChooseButton, c);
        gridBagNewLine(c);
        gridBagAdder(detailFileTopPanel, new JLabel("Glyph name: "), c);
        glyphNameField = new JTextField(20);
        textFieldOnChange(glyphNameField, text -> stateManager.setGlyphName(text));
        c.fill = GridBagConstraints.HORIZONTAL;
        gridBagAdder(detailFileTopPanel, glyphNameField, c);
        c.fill = 0;
        JButton chooseButton = new JButton("Open");
        chooseButton.addActionListener(e -> {
            loadDialog(detailFilePanel, stateManager.getDirectory(), null, selectedFile -> {
                String parent = selectedFile.getParent();
                dirTextField.setText(parent);
                dirTextField.postActionEvent();
                for (Preview sibling : siblings) {
                    sibling.previewStateManager.setDirectory(parent);
                }
                glyphNameField.setText(selectedFile.getName().replaceAll("[.]glyph\\z", ""));
                glyphNameField.postActionEvent();
                stateManager.loadAction();
            });
        });
        gridBagAdder(detailFileTopPanel, chooseButton, c);
        detailFilePanel.add(detailFileTopPanel);
        Box line3 = Box.createHorizontalBox();
        JButton saveButton = new JButton("Save");
        line3.add(saveButton);
        saveButton.addActionListener(e -> {
            boolean go = true;
            File file = stateManager.getFile();
            if (file.exists()) {
                File lastFile = stateManager.getLastFile();
                try {
                    if (lastFile == null || !Files.isSameFile(lastFile.toPath(), file.toPath())) {
                        if (JOptionPane.OK_OPTION != JOptionPane.showConfirmDialog(
                                detailFilePanel,
                                "Overwrite file?",
                                "Confirm", JOptionPane.OK_CANCEL_OPTION)) {
                            go = false;
                        }
                    }
                } catch (IOException e1) {
                    throw new RuntimeException(e1);
                }
            }
            if (go) {
                stateManager.saveAction();
            }
        });
        int width = 10;
        line3.add(Box.createHorizontalStrut(width));
        JButton loadButton = new JButton("Load");
        line3.add(loadButton);
        loadButton.addActionListener(e -> {
            if (JOptionPane.OK_OPTION == JOptionPane.showConfirmDialog(detailFilePanel, "Load glyph and clear undo history?", "Confirm", JOptionPane.OK_CANCEL_OPTION)) {
                stateManager.loadAction();
            }
        });
        line3.add(Box.createHorizontalStrut(width));
        JButton undoButton = new JButton("Undo history");
        undoButton.addActionListener(e -> {
            historyLister.showWindow();
        });
        line3.add(undoButton);
        line3.add(Box.createHorizontalStrut(width));
        JButton clearButton = new JButton("Clear");
        clearButton.addActionListener(e -> {
            if (JOptionPane.OK_OPTION == JOptionPane.showConfirmDialog(detailFilePanel, "Clear current drawing and undo history?", "Confirm", JOptionPane.OK_CANCEL_OPTION)) {
                stateManager.clear(true);
                glyphNameField.setText("");
                glyphNameField.postActionEvent();
            }
        });
        line3.add(clearButton);
        line3.add(Box.createHorizontalStrut(width));
        JButton composeButton = new JButton("Compose");
        composeButton.addActionListener(e -> {
            compositor.setVisible(true);
        });
        line3.add(composeButton);
        detailFilePanel.add(line3);
        siblings = new ArrayList<>();
        JPanel siblingsPanel = new JPanel();
        siblingsPanel.setLayout(new GridBagLayout());
        c = gridBagInit();
        gridBagAdder(siblingsPanel, new JLabel("Glyph name"), c);
        gridBagAdder(siblingsPanel, new JLabel("X shift"), c);
        gridBagAdder(siblingsPanel, new JLabel(""), c);
        gridBagAdder(siblingsPanel, new JLabel(""), c);
        gridBagAdder(siblingsPanel, new JLabel("Sel"), c);
        gridBagAdder(siblingsPanel, new JLabel(""), c);
        gridBagAdder(siblingsPanel, new JLabel(""), c);
        gridBagNewLine(c);
        for (int a = 0; a < 4; a++) {
            addSiblingsLine(siblingsPanel, c);
        }
        detailFilePanel.add(siblingsPanel);
    }

    private void loadDialog(Component parent, String directory, FileSystemView fsv, Consumer<File> onApprove) {
        JFileChooser jFileChooser = fsv == null ? new JFileChooser(new File(directory)) : new JFileChooser(new File(directory), fsv);
        JPanel previewPanel = new JPanel();
        Preview preview = new Preview();
        JPanel previewDrawPanel = new JPanel() {
            @Override
            protected void paintComponent(Graphics g) {
                super.paintComponent(g);
                preview.previewGlyphPainter.paintOnGraphics(this.getSize(), g, preview.previewGlyphDesignerModel.glyph, preview.previewGrid, null);
            }

        };
        previewDrawPanel.setMinimumSize(new Dimension(300, 300));
        previewDrawPanel.setPreferredSize(new Dimension(300, 300));
        previewPanel.add(previewDrawPanel);
        jFileChooser.addPropertyChangeListener(pce -> {
            if (pce.getPropertyName().equals(JFileChooser.DIRECTORY_CHANGED_PROPERTY)) {
                preview.previewStateManager.clear(true);
            } else if (pce.getPropertyName().equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY)) {
                File selectedFile = jFileChooser.getSelectedFile();
                if (selectedFile != null) {
                    preview.previewStateManager.setDirectory(selectedFile.getAbsoluteFile().getParent());
                    preview.previewStateManager.setGlyphName(selectedFile.getName().replaceAll(Pattern.quote(StateManager.FILE_EXT_GLYPH) + "\\z", ""));
                    preview.previewStateManager.loadAction();
                    preview.previewGlyphDesignerModel.selectCurveNone();
                    preview.previewGlyphPainter.forceRedraw();
                    previewDrawPanel.repaint();
                }
            }
        });
        jFileChooser.setAccessory(previewPanel);
        jFileChooser.setFileFilter(new FileFilter() {
            @Override
            public boolean accept(File f) {
                return f.isDirectory() || f.getName().endsWith(StateManager.FILE_EXT_GLYPH);
            }

            @Override
            public String getDescription() {
                return StateManager.FILE_EXT_GLYPH_DESCRIPTION;
            }
        });
        int result = jFileChooser.showDialog(parent, "Load");
        if (result == JFileChooser.APPROVE_OPTION) {
            File selectedFile = jFileChooser.getSelectedFile();
            onApprove.accept(selectedFile);
        }
    }

    class Preview {
        public ViewZoom viewZoom = new ViewZoom();
        public GlyphPainter previewGlyphPainter = new GlyphPainter(viewZoom);
        public GlyphDesignerModel previewGlyphDesignerModel = new GlyphDesignerModel();
        public Grid previewGrid = new Grid();
        public StateManager previewStateManager = new StateManager(previewGlyphDesignerModel, previewGrid);
        public double xShift = 0d, yShift = 0d;
        public JTextField glyphNameTextField;
        public JTextField glyphXShift;

        {
            viewZoom.getUpdateEventQueue().addEventConsumer(e -> previewGlyphPainter.invalidateCache());
            previewGlyphDesignerModel.getChangesToLayout().addEventConsumer(e -> repaintDrawPanel());
            previewGlyphDesignerModel.getChangesToCurves().addEventConsumer(e -> repaintDrawPanel());
        }

        public void updateViewForShiftByMasterView(ViewZoom masterView) {
            viewZoom.setAll(masterView.getX1() - xShift, masterView.getX2() - xShift, masterView.getY1() - yShift, masterView.getY2() - yShift);
        }
    }

    private void addSiblingsLine(JPanel siblingsPanel, GridBagConstraints c) {
        Preview preview = new Preview();
        JTextField glyphNameTextField = preview.glyphNameTextField = new JTextField("", 10);
        textFieldOnChange(glyphNameTextField, text -> preview.previewStateManager.setGlyphName(text));
        gridBagAdder(siblingsPanel, glyphNameTextField, c);
        JTextField glyphXShift = preview.glyphXShift = new JTextField("0", 8);
        textFieldOnChange(glyphXShift, text -> {
            preview.xShift = parseDoubleOrFraction(text, 0d);
            repaintDrawPanel();
        });
        gridBagAdder(siblingsPanel, glyphXShift, c);
        siblings.add(preview);
        JCheckBox selCheckbox = new JCheckBox();
        JButton loadButton = smallButton(new JButton("Load"));
        loadButton.addActionListener(e -> {
            preview.previewStateManager.loadAction();
            preview.previewGlyphDesignerModel.selectCurveNone();
            selCheckbox.setSelected(false);
        });
        gridBagAdder(siblingsPanel, loadButton, c);
        JButton chooseButton = smallButton(new JButton("Open"));
        chooseButton.addActionListener(e -> {
            loadDialog(siblingsPanel, stateManager.getDirectory(), new FixedDirFileSystemView(new File(stateManager.getDirectory())), selectedFile -> {
                glyphNameTextField.setText(selectedFile.getName().replaceAll("[.]glyph\\z", ""));
                glyphNameTextField.postActionEvent();
                preview.previewStateManager.loadAction();
                preview.previewGlyphDesignerModel.selectCurveNone();
                selCheckbox.setSelected(false);
            });
        });
        gridBagAdder(siblingsPanel, chooseButton, c);
        gridBagAdder(siblingsPanel, selCheckbox, c);
        selCheckbox.addActionListener(e -> {
            if (selCheckbox.isSelected()) {
                preview.previewGlyphDesignerModel.selectCurveAll();
            } else {
                preview.previewGlyphDesignerModel.selectCurveNone();
            }
        });
        JButton clearButton = smallButton(new JButton("Clear"));
        clearButton.addActionListener(e -> {
            preview.previewGlyphDesignerModel.clear();
            glyphNameTextField.setText("");
            glyphNameTextField.postActionEvent();
        });
        gridBagAdder(siblingsPanel, clearButton, c);
        JButton editButton = smallButton(new JButton("Edit"));
        editButton.addActionListener(e -> {
            if (!preview.previewStateManager.isFileEmpty()) {
                File file = preview.previewStateManager.getFile();
                int confirm = JOptionPane.showConfirmDialog(
                        siblingsPanel,
                        "Save current glyph before editing " + file,
                        "Confirm", JOptionPane.YES_NO_CANCEL_OPTION);
                if (JOptionPane.YES_OPTION == confirm) {
                    stateManager.saveAction();
                }
                if (JOptionPane.CANCEL_OPTION != confirm) {
                    String previousEditGlyphName = stateManager.getGlyphName();
                    double xShift = preview.xShift;
                    double yShift = preview.yShift;
                    String newEditGlyphName = preview.previewStateManager.getGlyphName();
                    stateManager.setGlyphName(newEditGlyphName);
                    glyphNameField.setText(newEditGlyphName);
                    stateManager.loadAction();
                    preview.xShift = 0;
                    preview.yShift = 0;
                    preview.previewStateManager.setGlyphName(previousEditGlyphName);
                    preview.glyphNameTextField.setText(previousEditGlyphName);
                    preview.previewStateManager.loadAction();
                    for (Preview sibling : siblings) {
                        sibling.xShift -= xShift;
                        sibling.glyphXShift.setText("" + sibling.xShift);
                        sibling.yShift -= yShift;
                    }

                }

            }
        });
        gridBagAdder(siblingsPanel, editButton, c);
        gridBagNewLine(c);
    }

    public static Double parseDoubleOrFraction(String text, Double alternative) {
        text = text.trim();
        try {
            if (text.startsWith("1/")) {
                return 1 / Double.parseDouble(text.substring(2));
            } else if (text.startsWith("-1/")) {
                return -1 / Double.parseDouble(text.substring(3));
            } else {
                return Double.parseDouble(text);
            }
        } catch (NumberFormatException nfe) {
            return alternative;
        }
    }

    private JButton smallButton(JButton button) {
        button.setMargin(new Insets(1, 1, 1, 1));
        return button;
    }

    private void textFieldOnChange(JTextField textField, Consumer<String> consume) {
        textField.addActionListener(e -> consume.accept(textField.getText()));
        textField.addFocusListener(new FocusListener() {
            @Override
            public void focusGained(FocusEvent e) {
            }

            @Override
            public void focusLost(FocusEvent e) {
                consume.accept(textField.getText());
            }
        });
    }

    public GridBagConstraints gridBagInit() {
        GridBagConstraints c = new GridBagConstraints();
        c.gridx = 0;
        c.gridy = 0;
        c.insets = new Insets(1, 2, 2, 1);
        return c;
    }

    public void gridBagAdder(JPanel gridBag, JComponent add, GridBagConstraints c) {
        gridBag.add(add, c);
        c.gridx++;
    }

    public void gridBagNewLine(GridBagConstraints c) {
        c.gridy++;
        c.gridx = 0;
    }

    private void detailZoomComponents(JPanel detailZoomPanel) {
        detailZoomPanel.setLayout(new BoxLayout(detailZoomPanel, BoxLayout.Y_AXIS));
        JPanel zoomEntryGrid = new JPanel();
        GridBagConstraints c = gridBagInit();
        zoomEntryGrid.setLayout(new GridBagLayout());
        gridBagAdder(zoomEntryGrid, new JLabel("X1, X2: "), c);
        x1ZoomTextField = new JTextField(8);
        enterToConfirmToolTip(x1ZoomTextField);
        x1ZoomTextField.addActionListener(new DoubleTextFieldActionListener(x1ZoomTextField, viewZoom::getX1, viewZoom::setX1));
        gridBagAdder(zoomEntryGrid, x1ZoomTextField, c);
        x2ZoomTextField = new JTextField(8);
        enterToConfirmToolTip(x2ZoomTextField);
        x2ZoomTextField.addActionListener(new DoubleTextFieldActionListener(x2ZoomTextField, viewZoom::getX2, viewZoom::setX2));
        gridBagAdder(zoomEntryGrid, x2ZoomTextField, c);
        gridBagAdder(zoomEntryGrid, getViewportMovePointsButton("←", "Move viewport left (CTRL for smaller step)", 0, -1), c);
        gridBagAdder(zoomEntryGrid, getViewportMovePointsButton("→", "Move viewport right (CTRL for smaller step)", 0, 1), c);
        gridBagNewLine(c);
        gridBagAdder(zoomEntryGrid, new JLabel("Y1, Y2: "), c);
        y1ZoomTextField = new JTextField(8);
        enterToConfirmToolTip(y1ZoomTextField);
        y1ZoomTextField.addActionListener(new DoubleTextFieldActionListener(y1ZoomTextField, viewZoom::getY1, viewZoom::setY1));
        gridBagAdder(zoomEntryGrid, y1ZoomTextField, c);
        y2ZoomTextField = new JTextField(8);
        enterToConfirmToolTip(y2ZoomTextField);
        y2ZoomTextField.addActionListener(new DoubleTextFieldActionListener(y2ZoomTextField, viewZoom::getY2, viewZoom::setY2));
        gridBagAdder(zoomEntryGrid, y2ZoomTextField, c);
        gridBagAdder(zoomEntryGrid, getViewportMovePointsButton("↑", "Move viewport up (CTRL for smaller step)", 1, 0), c);
        gridBagAdder(zoomEntryGrid, getViewportMovePointsButton("↓", "Move viewport down (CTRL for smaller step)", -1, 0), c);
        detailZoomPanel.add(zoomEntryGrid);
        Box line3 = Box.createHorizontalBox();
        JButton zoomDefault = new JButton("1x");
        zoomDefault.addActionListener(e -> viewZoom.setAll(0, 1, 0, 1));
        line3.add(zoomDefault);
        for (int i = 2; i <= 16; i *= 2) {
            line3.add(Box.createHorizontalStrut(6));
            JButton zoomPoint2 = new JButton("" + i + "x");
            final double halfView = 0.5 / i;
            zoomPoint2.addActionListener(e -> {
                List<Point> points = glyphDesignerModel.getSelectedPointOnlyStream().collect(Collectors.toList());
                if (points.isEmpty()) {
                    // try all points of selected curves
                    points = glyphDesignerModel.getSelectedCurveStream().flatMap(selc -> selc.allPoints().stream()).collect(Collectors.toList());
                }
                if (!points.isEmpty()) {
                    double x = 0, y = 0;
                    for (Point p : points) {
                        x += p.x;
                        y += p.y;
                    }
                    x /= points.size();
                    y /= points.size();
                    viewZoom.setAll(viewRound(x - halfView), viewRound(x + halfView), viewRound(y - halfView), viewRound(y + halfView));
                }
            });
            line3.add(zoomPoint2);
        }
        JButton zoomPlus = new JButton("+");
        zoomPlus.addActionListener(e -> viewZoom.magnify((e.getModifiers() & ActionEvent.CTRL_MASK) != 0 ? 1.25 : 2));
        zoomPlus.setToolTipText("CTRL for smaller step");
        line3.add(zoomPlus);
        JButton zoomMinus = new JButton("-");
        zoomMinus.addActionListener(e -> viewZoom.magnify((e.getModifiers() & ActionEvent.CTRL_MASK) != 0 ? 0.8 : 0.5));
        zoomMinus.setToolTipText("CTRL for smaller step");
        line3.add(zoomMinus);
        detailZoomPanel.add(line3);
        viewZoom.getUpdateEventQueue().addEventConsumer(e -> {
            x1ZoomTextField.setText("" + viewZoom.getX1());
            y1ZoomTextField.setText("" + viewZoom.getY1());
            x2ZoomTextField.setText("" + viewZoom.getX2());
            y2ZoomTextField.setText("" + viewZoom.getY2());
        });
    }

    private JButton getViewportMovePointsButton(String label, String toolTip, int my, int mx) {
        JButton button = new JButton(label);
        button.setToolTipText(toolTip);
        button.addActionListener(e -> viewZoom.getUpdateEventQueue().mergeEventsDuring(() -> {
            int div = (e.getModifiers() & ActionEvent.CTRL_MASK) != 0 ? 16 : 4;
            double w = (viewZoom.getX2() - viewZoom.getX1()) / div;
            double h = (viewZoom.getY2() - viewZoom.getY1()) / div;
            viewZoom.setX1(viewRound(viewZoom.getX1() + w * mx));
            viewZoom.setX2(viewRound(viewZoom.getX2() + w * mx));
            viewZoom.setY1(viewRound(viewZoom.getY1() + h * my));
            viewZoom.setY2(viewRound(viewZoom.getY2() + h * my));
        }));
        button.setPreferredSize(new Dimension(50, button.getPreferredSize().height));
        return button;
    }

    public static final double VIEW_ROUNDING = 1e5;

    private double viewRound(double x) {
        return Math.round(x * VIEW_ROUNDING) / VIEW_ROUNDING;
    }

    private boolean avoidRecursiveSelection;

    private void curveListComponents(JPanel curveListPanel) {
        curveListPanel.setLayout(new BoxLayout(curveListPanel, BoxLayout.Y_AXIS));

        GlyphDesignerModelListModel gdmListModel = new GlyphDesignerModelListModel(glyphDesignerModel);
        curveList = new JList<>(gdmListModel);
        curveList.setLayoutOrientation(JList.VERTICAL_WRAP);

        this.glyphDesignerModel.getChangesToCurves().addEventConsumer((obj) -> {
            repaintDrawPanel();
            if (!avoidRecursiveSelection) {
                try {
                    avoidRecursiveSelection = true;
                    ArrayList<Curve> curves = glyphDesignerModel.glyph.curves;
                    curveList.clearSelection();
                    for (int i = 0; i < curves.size(); i++) {
                        Curve curve = curves.get(i);
                        if (curve.selected) {
                            curveList.addSelectionInterval(i, i);
                        }
                    }
                } finally {
                    avoidRecursiveSelection = false;
                }
            }
        });
        this.glyphDesignerModel.getChangesToLayout().addEventConsumer((obj) -> {
            if (obj instanceof GlyphDesignerModel.Added) {
                int index = ((GlyphDesignerModel.Added) obj).getIndex();
                gdmListModel.fireIntervalAdded(gdmListModel, index, index);
            } else if (obj instanceof GlyphDesignerModel.Removed) {
                int index = ((GlyphDesignerModel.Removed) obj).getIndex();
                gdmListModel.fireIntervalRemoved(gdmListModel, index, index);
            } else if (obj instanceof GlyphDesignerModel.MultiRemoved) {
                gdmListModel.fireContentsChanged(gdmListModel, 0, gdmListModel.getSize() - 1);
            } else {
                gdmListModel.fireContentsChanged(gdmListModel, 0, gdmListModel.getSize() - 1);
            }
            repaintDrawPanel();
        });

        curveList.addListSelectionListener(e -> {
            int[] selectedIndices = curveList.getSelectedIndices();
            if (!avoidRecursiveSelection) {
                try {
                    avoidRecursiveSelection = true;
                    glyphDesignerModel.selectCurveMulti(selectedIndices);
                } finally {
                    avoidRecursiveSelection = false;
                }
            }
        });
        JButton removeButton = new JButton("Remove (Del)");
        AbstractAction delAction = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (shortcutOutsideOfTypingControls()) {
                    glyphDesignerModel.removeCurves();
                    stateManager.addLabel("Removed curve");
                }
            }
        };
        drawPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0), "DEL");
        drawPanel.getActionMap().put("DEL", delAction);
        removeButton.addActionListener(delAction);

        JButton newButton = new JButton("New (N)");
        AbstractAction newAction = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (shortcutOutsideOfTypingControls()) {
                    curveList.clearSelection();
                    glyphDesignerModel.newCurve();
                    stateManager.addLabel("New curve");
                }
            }
        };
        drawPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("N"), "N");
        drawPanel.getActionMap().put("N", newAction);
        newButton.addActionListener(newAction);

        Box line1 = Box.createHorizontalBox();
        curveListPanel.add(line1);
        line1.add(newButton);
        line1.add(Box.createHorizontalGlue());
        line1.add(removeButton);
        Box line2 = Box.createHorizontalBox();
        curveListPanel.add(line2);

        line2.add(curveList);

        // selections
        // select all
        AbstractAction selectAllAction = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (shortcutOutsideOfTypingControls()) {
                    curveList.setSelectionInterval(0, curveList.getModel().getSize() - 1);
                    stateManager.addLabel("Selected all");
                }
            }
        };
        drawPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_A, KeyEvent.CTRL_DOWN_MASK), "SelectAll");
        drawPanel.getActionMap().put("SelectAll", selectAllAction);

        // unselect
        AbstractAction unselectAction = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (shortcutOutsideOfTypingControls()) {
                    curveList.clearSelection();
                    stateManager.addLabel("Unselected");
                }
            }
        };
        drawPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0), "unselect");
        drawPanel.getActionMap().put("unselect", unselectAction);

    }

    private boolean shortcutOutsideOfTypingControls() {
        return !(KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner() instanceof JTextField);
    }

    private void selectedPointsComponents(JPanel selectedPointsPanel) {
        selectedPointsPanel.setLayout(new BoxLayout(selectedPointsPanel, BoxLayout.Y_AXIS));
        Box line1 = Box.createHorizontalBox();
        line1.add(new JLabel("Move by "));
        JTextField moveByTextField = new JTextField("1/10");
        line1.add(Box.createHorizontalStrut(6));
        line1.add(moveByTextField);
        line1.add(Box.createHorizontalStrut(6));
        Supplier<Double> moveByReader = () -> parseDoubleOrFraction(moveByTextField.getText(), null);
        addMovePointsButton(line1, moveByReader, "↑", "Move up by specified step", 1, 0);
        addMovePointsButton(line1, moveByReader, "↓", "Move down by specified step", -1, 0);
        addMovePointsButton(line1, moveByReader, "←", "Move left by specified step", 0, -1);
        addMovePointsButton(line1, moveByReader, "→", "Move right by specified step", 0, 1);
        selectedPointsPanel.add(line1);

        Box line2 = Box.createHorizontalBox();
        addActionButton(line2, "⇈", "Align Y coordinates to top", glyphDesignerModel::alignPointsTop);
        addActionButton(line2, "⇊", "Align Y coordinates to bottom", glyphDesignerModel::alignPointsBottom);
        addActionButton(line2, "⇇", "Align X coordinates to left", glyphDesignerModel::alignPointsLeft);
        addActionButton(line2, "⇉", "Align X coordinates to right", glyphDesignerModel::alignPointsRight);
        selectedPointsPanel.add(line2);
    }

    private void selectedCurvesComponents(JPanel detailCurvePanel) {
        detailCurvePanel.setLayout(new BoxLayout(detailCurvePanel, BoxLayout.Y_AXIS));

        Box line1 = Box.createHorizontalBox();
        line1.add(new JLabel("Move by "));
        JTextField moveByTextField = new JTextField("1/10");
        Supplier<Double> moveByReader = () -> parseDoubleOrFraction(moveByTextField.getText(), null);
        line1.add(Box.createHorizontalStrut(6));
        line1.add(moveByTextField);
        line1.add(Box.createHorizontalStrut(6));
        addMoveButton(line1, moveByReader, "↑", "Move up by specified step", 1, 0);
        addMoveButton(line1, moveByReader, "↓", "Move down by specified step", -1, 0);
        addMoveButton(line1, moveByReader, "←", "Move left by specified step", 0, -1);
        addMoveButton(line1, moveByReader, "→", "Move right by specified step", 0, 1);
        detailCurvePanel.add(line1);


        Box line2 = Box.createHorizontalBox();
        addActionButton(line2, "↶", "Rotate CCW", glyphDesignerModel::rotateCCW);
        addActionButton(line2, "↷", "Rotate CW", glyphDesignerModel::rotateCW);
        addActionButton(line2, "↔", "Flip left-right", glyphDesignerModel::flipLeftRight);
        addActionButton(line2, "↕", "Flip top-bottom", glyphDesignerModel::flipTopBottom);
        detailCurvePanel.add(line2);

        AbstractAction flipDirectionAction = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (shortcutOutsideOfTypingControls()) {
                    glyphDesignerModel.flipDirection();
                }
            }
        };
        drawPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("D"), "D");
        drawPanel.getActionMap().put("D", flipDirectionAction);

        Box line3 = Box.createHorizontalBox();
        addActionButton(line3, "P1↔P4 (D)", "Flip direction", () -> flipDirectionAction.actionPerformed(null));
        addActionButton(line3, "⏆ 1/4", "Split at 1/4", glyphDesignerModel::split1div4);
        addActionButton(line3, "⏆ 1/2", "Split at 1/2", glyphDesignerModel::split2div4);
        addActionButton(line3, "⏆ 3/4", "Split at 3/4", glyphDesignerModel::split3div4);
        addActionButton(line3, "Line", "Make lines from curves", glyphDesignerModel::makeLine);
        detailCurvePanel.add(line3);


        AbstractAction copyAction = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if (shortcutOutsideOfTypingControls()) {
                    String s = glyphDesignerModel.serializeSelection();
                    StringSelection selection = new StringSelection(s);
                    Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
                    clipboard.setContents(selection, selection);
                }
            }
        };
        drawPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_DOWN_MASK), "Copy");
        drawPanel.getActionMap().put("Copy", copyAction);

        AbstractAction pasteAction = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent ev) {
                if (shortcutOutsideOfTypingControls()) {
                    Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
                    try {
                        String s = (String) clipboard.getData(DataFlavor.stringFlavor);
                        int added = glyphDesignerModel.pasteSerialized(s);
                        curveList.clearSelection();
                        int size = glyphDesignerModel.glyph.curves.size();
                        curveList.addSelectionInterval(size - added, size);
                        repaintDrawPanel();
                    } catch (UnsupportedFlavorException | IOException e) {
                        throw new RuntimeException(e);
                    }
                    stateManager.addLabel("Pasted from clipboard");
                }
            }
        };
        drawPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_V, KeyEvent.CTRL_DOWN_MASK), "Paste");
        drawPanel.getActionMap().put("Paste", pasteAction);

        Box line4 = Box.createHorizontalBox();
        JButton copyButton = new JButton("Copy (Ctrl-C)");
        copyButton.addActionListener(copyAction);
        line4.add(copyButton);

        JButton pasteButton = new JButton("Paste (Ctrl-V)");
        pasteButton.addActionListener(pasteAction);
        line4.add(pasteButton);

        JButton copyAsNumberButton = new JButton("Copy as number");
        copyAsNumberButton.addActionListener(e -> {
            String s = glyphDesignerModel.serializeSelectionAsNumber();
            StringSelection selection = new StringSelection(s);
            Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
            clipboard.setContents(selection, selection);
        });
        line4.add(copyAsNumberButton);
        detailCurvePanel.add(line4);
    }

    private void addActionButton(JComponent line, String label, String toolTip, Runnable modelAction) {
        JButton actionButton = new JButton(label);
        actionButton.setMinimumSize(new Dimension(60, actionButton.getMinimumSize().height));
        actionButton.addActionListener(e -> {
            modelAction.run();
            stateManager.addLabel(toolTip);
        });
        actionButton.setToolTipText(toolTip);
        line.add(actionButton);
        line.add(Box.createHorizontalStrut(10));
    }

    private void addMoveButton(JComponent line, Supplier<Double> moveByReader, String label, String toolTip, int my, int mx) {
        JButton moveButton = new JButton(label);
        moveButton.setToolTipText(toolTip);
        moveButton.addActionListener(e -> {
            Double step = moveByReader.get();
            if (step != null) {
                glyphDesignerModel.moveRel(step * mx, step * my);
                stateManager.addLabel(toolTip);
            }
        });
        moveButton.setMinimumSize(new Dimension(50, moveButton.getMinimumSize().height));
        line.add(moveButton);
        line.add(Box.createHorizontalStrut(6));
    }

    private void addMovePointsButton(JComponent line, Supplier<Double> moveByReader, String label, String toolTip, int my, int mx) {
        JButton moveButton = new JButton(label);
        moveButton.setToolTipText(toolTip);
        moveButton.addActionListener(e -> {
            Double step = moveByReader.get();
            if (step != null) {
                glyphDesignerModel.movePoints(step * mx, step * my);
                stateManager.addLabel(toolTip);
            }
        });
        moveButton.setMinimumSize(new Dimension(50, moveButton.getMinimumSize().height));
        line.add(moveButton);
        line.add(Box.createHorizontalStrut(6));
    }


    private void detailGridComponents(JPanel detailGridPanel) {
        JPanel gridPanel = new JPanel();
        detailGridPanel.setLayout(new BoxLayout(detailGridPanel, BoxLayout.Y_AXIS));
        detailGridPanel.add(gridPanel);
        int count = grid.getxGridDivisions().length;
        gridPanel.setLayout(new GridLayout(3, count + 1));
        gridPanel.add(new JLabel("X divs: "));
        xGridDivisionsTextField = new JTextField[count];
        for (int i = 0; i < count; i++) {
            xGridDivisionsTextField[i] = new JTextField("" + grid.getxGridDivisions()[i], 3);
            final int j = i;
            GridItemTextFieldActionListener listener = new GridItemTextFieldActionListener(
                    xGridDivisionsTextField[j],
                    () -> grid.getxGridDivisions()[j],
                    v -> grid.setxGridDivisions(j, v));
            xGridDivisionsTextField[i].addActionListener(listener);
            xGridDivisionsTextField[i].addFocusListener(listener);
            gridPanel.add(xGridDivisionsTextField[j]);
        }
        gridPanel.add(new JLabel("Y divs: "));
        yGridDivisionsTextField = new JTextField[count];
        for (int i = 0; i < count; i++) {
            yGridDivisionsTextField[i] = new JTextField("" + grid.getyGridDivisions()[i], 3);
            final int j = i;
            GridItemTextFieldActionListener listener = new GridItemTextFieldActionListener(
                    yGridDivisionsTextField[j],
                    () -> grid.getyGridDivisions()[j],
                    v -> grid.setyGridDivisions(j, v));
            yGridDivisionsTextField[i].addActionListener(listener);
            yGridDivisionsTextField[i].addFocusListener(listener);
            gridPanel.add(yGridDivisionsTextField[j]);
        }
        activeGridGroup = new ButtonGroup();
        activeGridRadioButton = new JRadioButton[count + 1];
        JRadioButton gridOffButton = null;
        for (int i = -1; i < count; i++) {
            JRadioButton radioButton = new JRadioButton(i < 0 ? "off" : "");
            activeGridGroup.add(radioButton);
            if (i >= 0) {
                activeGridRadioButton[i] = radioButton;
            } else {
                gridOffButton = radioButton;
            }
            final int j = i;
            radioButton.addActionListener(e -> grid.setActiveGrid(j));
            gridPanel.add(radioButton);
        }
        final JRadioButton gridOffButtonFinal = gridOffButton;
        JCheckBox xyGuidesCheckBox = new JCheckBox("X, Y Guides");
        detailGridPanel.add(xyGuidesCheckBox);
        xyGuidesCheckBox.addActionListener(e -> grid.setGuidesActive(xyGuidesCheckBox.isSelected()));
        JTextField xGuides = new JTextField();
        String toolTipText = "Use decimals or fractions, separate multiple values by space";
        xGuides.setToolTipText(toolTipText);
        detailGridPanel.add(xGuides);
        xGuides.addActionListener(e -> grid.setxGuidesString(xGuides.getText()));
        JTextField yGuides = new JTextField();
        yGuides.setToolTipText(toolTipText);
        detailGridPanel.add(yGuides);
        yGuides.addActionListener(e -> grid.setyGuidesString(yGuides.getText()));
        grid.getImportEventQueue().addEventConsumer(e -> {
            xGuides.setText(grid.getxGuidesString());
            yGuides.setText(grid.getyGuidesString());
            for (int i = 0; i < grid.getxGridDivisions().length; i++) {
                xGridDivisionsTextField[i].setText("" + grid.getxGridDivisions()[i]);
                yGridDivisionsTextField[i].setText("" + grid.getyGridDivisions()[i]);
            }
            int activeGrid = grid.getActiveGrid();
            if (activeGrid == -1) {
                gridOffButtonFinal.getModel().setSelected(true);
            } else {
                activeGridRadioButton[grid.getActiveGrid()].getModel().setSelected(true);
            }
            xyGuidesCheckBox.setSelected(grid.isGuidesActive());
        });
    }

    private void enterToConfirmToolTip(JTextField textField) {
        textField.setToolTipText("press enter to confirm");
    }

    private void updateDetailPointUI() {
        userUpdatesPoints = false;
        try {
            List<Point> points = glyphDesignerModel.getSelectedPointOnlyStream().collect(Collectors.toList());
            if (points.size() >= 1) {
                Point p = points.get(0);
                ArrayList<Curve> curves = glyphDesignerModel.glyph.curves;
                String pName = null;
                String cName = null;
                for (int i = 0; i < curves.size(); i++) {
                    Curve curve = curves.get(i);
                    if (curve.selected) {
                        pName = curve.describePoint(p);
                        if (pName != null) {
                            cName = "Curve " + (i + 1);
                            break;
                        }
                    }
                }
                if (points.size() == 2) {
                    detailXTextField.setEnabled(false);
                    detailXTextField.setText("");
                    detailYTextField.setEnabled(false);
                    detailYTextField.setText("");
                    Point p1 = points.get(0);
                    Point p4 = points.get(1);
                    double dx = p4.x - p1.x;
                    double dy = p4.y - p1.y;
                    double angle = Math.atan2(dy, dx) / Math.PI * 180;
                    double len = Math.sqrt(dx * dx + dy * dy);
                    detailInfoLabel.setText(String.format("Two points selected. Angle %.2f deg, len %.6f, dx %.3f, dy %.3f.", angle, len, dx, dy));
                } else if (points.size() > 2) {
                    detailInfoLabel.setText("Multiple points selected (" + points.size() + ")");
                    detailXTextField.setEnabled(false);
                    detailXTextField.setText("");
                    detailYTextField.setEnabled(false);
                    detailYTextField.setText("");
                } else {
                    detailInfoLabel.setText("Point " + pName + " in " + cName);
                    detailXTextField.setText("" + p.x);
                    detailYTextField.setText("" + p.y);
                    detailXTextField.setEnabled(true);
                    detailYTextField.setEnabled(true);
                }
                // TODO: multiple points
                detailXTextField.setText("" + p.x);
                detailYTextField.setText("" + p.y);
                detailXTextField.setEnabled(true);
                detailYTextField.setEnabled(true);
            } else { // no selection, set defaults
                detailXTextField.setEnabled(false);
                detailXTextField.setText("");
                detailYTextField.setEnabled(false);
                detailYTextField.setText("");
                List<Curve> curveList = glyphDesignerModel.getSelectedCurveStream().collect(Collectors.toList());
                if (curveList.size() == 1) {
                    Curve curve = curveList.get(0);
                    Point p1 = curve.p1;
                    Point p4 = curve.p4;
                    double dx = p4.x - p1.x;
                    double dy = p4.y - p1.y;
                    double angle = Math.atan2(dy, dx) / Math.PI * 180;
                    double len = Math.sqrt(dx * dx + dy * dy);
                    detailInfoLabel.setText(String.format("Curve P1-P4 angle %.2f deg, len %.6f, dx %.3f, dy %.3f.", angle, len, dx, dy));
                } else {
                    detailInfoLabel.setText("No point selected.");
                }
            }
        } finally {
            userUpdatesPoints = true;
        }
    }


    private void detailPointComponents(JPanel detailPointPanel) {
        detailPointPanel.setLayout(new BoxLayout(detailPointPanel, BoxLayout.Y_AXIS));
        JPanel detailLine0 = new JPanel();
        detailLine0.setLayout(new BoxLayout(detailLine0, BoxLayout.X_AXIS));
        detailInfoLabel = new JLabel();
        detailLine0.add(detailInfoLabel);
        detailPointPanel.add(detailLine0);

        detailPointPanel.add(Box.createVerticalStrut(5));
        JPanel detailLine2 = new JPanel();
        detailLine2.setLayout(new BoxLayout(detailLine2, BoxLayout.X_AXIS));
        detailLine2.add(new JLabel("x: "));
        detailLine2.add(Box.createHorizontalStrut(3));
        detailXTextField = new JTextField(16);
        enterToConfirmToolTip(detailXTextField);
        detailXTextField.addActionListener(e -> {
            if (userUpdatesPoints) {
                try {
                    double x = Double.parseDouble(detailXTextField.getText());
                    glyphDesignerModel.getSelectedPointStream().forEach(p -> p.x = x);
                    glyphDesignerModel.getChangesToCurves().postEvent();
                } catch (NumberFormatException nfe) {
                    // ignore
                }
            }
        });
        detailLine2.add(detailXTextField);
        detailPointPanel.add(detailLine2);
        detailPointPanel.add(Box.createVerticalStrut(5));

        JPanel detailLine3 = new JPanel();
        detailLine3.setLayout(new BoxLayout(detailLine3, BoxLayout.X_AXIS));
        detailLine3.add(new JLabel("y: "));
        detailLine3.add(Box.createHorizontalStrut(3));
        detailYTextField = new JTextField(16);
        enterToConfirmToolTip(detailYTextField);
        detailYTextField.addActionListener(e -> {
            if (userUpdatesPoints) {
                try {
                    double y = Double.parseDouble(detailYTextField.getText());
                    glyphDesignerModel.getSelectedPointStream().forEach(p -> p.y = y);
                    glyphDesignerModel.getChangesToCurves().postEvent();
                } catch (NumberFormatException nfe) {
                    // ignore
                }
            }
        });
        detailLine3.add(detailYTextField);
        detailPointPanel.add(detailLine3);
        detailPointPanel.add(Box.createVerticalStrut(5));
    }

    private void repaintDrawPanel() {
        glyphPainter.forceRedraw();
        drawPanel.repaint();
    }

    private void repaintDrawPanelInvalidateCache() {
        glyphPainter.invalidateCache();
        drawPanel.repaint();
    }


    public static void main(String[] args) {
        new GlyphDesigner();
    }

    public static class GridItemTextFieldActionListener implements ActionListener, FocusListener {
        private final JTextField textField;
        private final IntSupplier getter;
        private final IntConsumer setter;

        public GridItemTextFieldActionListener(JTextField textField, IntSupplier getter, IntConsumer setter) {
            this.textField = textField;
            this.getter = getter;
            this.setter = setter;
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            update();
        }

        private void update() {
            String text = textField.getText();
            if (!text.isEmpty()) {
                try {
                    setter.accept(Integer.valueOf(text.trim()));
                } catch (NumberFormatException nfe) {
                }
            }
            int i = getter.getAsInt();
            textField.setText("" + i);
        }

        @Override
        public void focusGained(FocusEvent e) {
        }

        @Override
        public void focusLost(FocusEvent e) {
            update();
        }
    }

    public static class DoubleTextFieldActionListener implements ActionListener {
        private final JTextField textField;
        private final DoubleSupplier getter;
        private final DoubleConsumer setter;

        public DoubleTextFieldActionListener(JTextField textField, DoubleSupplier getter, DoubleConsumer setter) {
            this.textField = textField;
            this.getter = getter;
            this.setter = setter;
            textField.setText("" + getter.getAsDouble());
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            String text = textField.getText();
            try {
                setter.accept(Double.valueOf(text.trim()));
            } catch (NumberFormatException nfe) {
                double d = getter.getAsDouble();
                textField.setText("" + d);
            }
        }
    }

}
