/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import javax.xml.bind.annotation.XmlRootElement;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Character composed of multiple curves
 */
public class GlyphDesignerModel {
    final Glyph glyph = new Glyph();

    private GlyphDefaults glyphDefaults = new GlyphDefaults();

    private EventQueue changesToLayout = new EventQueue();
    private EventQueue changesToCurves = new EventQueue();
    private MousePos pressLocation;
    private MousePos precisePressLocation;
    private boolean pressBefore;
    private NewMode newMode;
    private MousePos rectangleSelection;
    private MousePos rectangleSelectionEnd;

    public MousePos getRectangleSelection() {
        return rectangleSelection;
    }

    public MousePos getRectangleSelectionEnd() {
        return rectangleSelectionEnd;
    }

    public void alignPointsTop() {
        Optional<Double> od = getSelectedPointOnlyStream().map(p -> p.y).max(Double::compareTo);
        od.ifPresent(d -> getSelectedPointOnlyStream().forEach(p -> p.y = d));
        changesToCurves.postEvent();
    }

    public void alignPointsBottom() {
        Optional<Double> od = getSelectedPointOnlyStream().map(p -> p.y).min(Double::compareTo);
        od.ifPresent(d -> getSelectedPointOnlyStream().forEach(p -> p.y = d));
        changesToCurves.postEvent();
    }

    public void alignPointsRight() {
        Optional<Double> od = getSelectedPointOnlyStream().map(p -> p.x).max(Double::compareTo);
        od.ifPresent(d -> getSelectedPointOnlyStream().forEach(p -> p.x = d));
        changesToCurves.postEvent();
    }

    public void alignPointsLeft() {
        Optional<Double> od = getSelectedPointOnlyStream().map(p -> p.x).min(Double::compareTo);
        od.ifPresent(d -> getSelectedPointOnlyStream().forEach(p -> p.x = d));
        changesToCurves.postEvent();
    }

    public void movePoints(double dx, double dy) {
        getSelectedPointOnlyStream().forEach(p -> p.moveRel(dx, dy));
        changesToCurves.postEvent();
    }

    public void clear() {
        glyph.curves.clear();
        changesToLayout.postEvent();
    }

    public void flipDirection() {
        getSelectedCurveStream().forEach(c -> {
            Curve c2 = c.duplicate();
            c.p1 = c2.p4;
            c.p2 = c2.p3;
            c.p3 = c2.p2;
            c.p4 = c2.p1;
        });
        changesToCurves.postEvent();
    }

    public void split1div4() {
        splitInto(new CurveSplitterDiv(1 / 4.0));
    }

    public void split2div4() {
        splitInto(new CurveSplitterDiv(1 / 2.0));
    }

    public void split3div4() {
        splitInto(new CurveSplitterDiv(3 / 4.0));
    }

    public void makeLine() {
        getSelectedCurveStream().forEach(c -> {
            Curve c2 = c.duplicate();
            c.p2 = c2.p1;
            c.p3 = c2.p4;
        });
        changesToCurves.postEvent();
    }

    @FunctionalInterface
    interface CurveSplitter {
        Stream<Curve> split(Curve c);
    }

    static class CurveSplitterDiv implements CurveSplitter {
        private final double position;

        CurveSplitterDiv(double position) {
            this.position = position;
        }

        @Override
        public Stream<Curve> split(Curve c) {
            double split1p1x = c.p1.x;
            double split1p1y = c.p1.y;
            double split2p4x = c.p4.x;
            double split2p4y = c.p4.y;
            double split1p2x = mid(c.p1.x, c.p2.x);
            double split1p2y = mid(c.p1.y, c.p2.y);
            double split2p3x = mid(c.p3.x, c.p4.x);
            double split2p3y = mid(c.p3.y, c.p4.y);
            double tmpx = mid(c.p2.x, c.p3.x);
            double tmpy = mid(c.p2.y, c.p3.y);
            double split1p3x = mid(split1p2x, tmpx);
            double split1p3y = mid(split1p2y, tmpy);
            double split2p2x = mid(tmpx, split2p3x);
            double split2p2y = mid(tmpy, split2p3y);
            double split1p4_2p1x = mid(split1p3x, split2p2x);
            double split1p4_2p1y = mid(split1p3y, split2p2y);
            return Stream.of(
                    new Curve(new Point(split1p1x, split1p1y), new Point(split1p2x, split1p2y), new Point(split1p3x, split1p3y), new Point(split1p4_2p1x, split1p4_2p1y), true),
                    new Curve(new Point(split1p4_2p1x, split1p4_2p1y), new Point(split2p2x, split2p2y), new Point(split2p3x, split2p3y), new Point(split2p4x, split2p4y), true));
        }

        private double mid(double a, double b) {
            return a + (b - a) * position;
        }
    }

    private void splitInto(CurveSplitter curveSplitter) {
        glyph.curves = glyph.curves.stream().flatMap(c -> {
            if (c.selected) {
                return curveSplitter.split(c);
            } else {
                return Stream.of(c);
            }
        }).collect(ArrayList::new, ArrayList::add, ArrayList::addAll);
        changesToCurves.postEvent();
        changesToLayout.postEvent();
    }


    public enum NewMode {
        expectingP1,
        expectingP4
    }


    public EventQueue getChangesToLayout() {
        return changesToLayout;
    }

    public EventQueue getChangesToCurves() {
        return changesToCurves;
    }

    public void selectCurveAll() {
        glyph.curves.forEach(Curve::select);
        changesToCurves.postEvent();
    }

    public void selectCurveNone() {
        glyph.curves.forEach(Curve::deselect);
        changesToCurves.postEvent();
    }

    public void selectCurveMulti(int... selectedIndices) {
        glyph.curves.forEach(Curve::deselect);
        ArrayList<Curve> curves = glyph.curves;
        for (int selectedIndex : selectedIndices) {
            if (selectedIndex >= 0 && selectedIndex < curves.size()) {
                curves.get(selectedIndex).select();

            }
        }
        changesToCurves.postEvent();
    }

    public void addCurveToSelection(int idx) {
        ArrayList<Curve> curves = glyph.curves;
        if (idx >= 0 && idx < curves.size()) {
            curves.get(idx).select();
        }
        changesToCurves.postEvent();
    }

    public void removeCurveFromSelection(int idx) {
        ArrayList<Curve> curves = glyph.curves;
        if (idx >= 0 && idx < curves.size()) {
            curves.get(idx).deselect();
        }
        changesToCurves.postEvent();
    }

    public int getSelectedIndex() {
        ArrayList<Curve> curves = glyph.curves;
        for (int i = 0; i < curves.size(); i++) {
            Curve curve = curves.get(i);
            if (curve.selected) {
                return i;
            }
        }
        return -1;
    }

    public void removeCurves() {
        glyph.curves.removeIf(curve -> curve.selected);
        changesToLayout.postEvent(new MultiRemoved());
    }

    public void newCurve() {
        newMode = NewMode.expectingP1;
        selectCurveNone();
        changesToLayout.postEvent();
    }

    private void handleNewStart(Grid grid, MousePos m) {
        newMode = NewMode.expectingP4;
        m = grid.clipToGrid(m);
        Curve c = new Curve();
        c.p1 = new Point(m.x, m.y);
        c.p2 = new Point(m.x, m.y);
        c.p3 = new Point(m.x, m.y);
        c.p4 = new Point(m.x, m.y);
        c.selected = true;
        glyph.curves.add(c);
        int index = glyph.curves.size() - 1;
        changesToLayout.postEvent(new Added(index));
        changesToCurves.postEvent();
    }

    private void handleNewCont(Grid grid, MousePos m) {
        m = grid.clipToGrid(m);
        Curve lastCurve = glyph.curves.get(glyph.curves.size() - 1);
        lastCurve.p3.moveTo(m);
        lastCurve.p4.moveTo(m);
        changesToLayout.postEvent();
        changesToCurves.postEvent();
    }

    private void handleNewFinish(Grid grid, MousePos m) {
        newMode = null;
        handleNewCont(grid, m);
    }

    public Stream<Point> getAllPointsInSelectedCurvesStream() {
        return getSelectedCurveStream().flatMap(c -> c.allPoints().stream());
    }


    public void handlePosition(Grid grid, MousePos m) {
        if (newMode == NewMode.expectingP4) {
            handleNewCont(grid, m);
        }
    }


    public void handleDrag(Grid grid, MousePos m) {
        if (m != null) {
            if (pressBefore) { // selections only when press precedes drag
                pressBefore = false;
                if (newMode == NewMode.expectingP1) {
                    handleNewStart(grid, m);
                    return;
                }
                List<PointAndDistance> closestPointsAndDistance = getClosestPointsAndDistance(m);
                Optional<PointAndDistance> first = closestPointsAndDistance.stream().filter(pd -> pd.p.selected).findFirst();
                if (first.isPresent()) { // drag can start on a selected point
                    MousePos mc = grid.clipToGrid(m);
                    Point pClosest = first.get().p;
                    double dx = mc.x - pClosest.x;
                    double dy = mc.y - pClosest.y;
                    getSelectedPointStream().forEach(p -> p.moveRel(dx, dy));
                    pressLocation = mc;
                    changesToCurves.postEvent();
                } else if (!closestPointsAndDistance.isEmpty()) {
                    getSelectedPointStream().forEach(p -> p.selected = false);
                    MousePos mc = grid.clipToGrid(m);
                    closestPointsAndDistance.forEach(pd -> {
                        pd.p.selected = true;
                        pd.p.moveTo(mc);
                    });
                    pressLocation = mc;
                    changesToCurves.postEvent();
                } else {
                    pressLocation = null;
                    rectangleSelection = m;
                }
            }
            if (newMode == NewMode.expectingP4) {
                handleNewCont(grid, m);
            }
            if (pressLocation != null) {
                handleDragOrFinish(grid, pressLocation, m);
            }
            if (rectangleSelection != null) {
                rectangleSelectionEnd = m;
                changesToCurves.postEvent();// ensure redraw
            }
        }
    }

    private void handleDragOrFinish(Grid grid, MousePos pressLocation, MousePos m) {
        MousePos mousePos = grid.clipToGrid(m);
        if (mousePos != null) {
            double dx = mousePos.x - pressLocation.x;
            double dy = mousePos.y - pressLocation.y;
            getSelectedPointStream().forEach(p -> p.moveRel(dx, dy));
            this.pressLocation = mousePos;
            getChangesToCurves().postEvent();
        }
    }

    public void handlePress(Grid grid, MousePos m) {
        pressLocation = grid.clipToGrid(m);
        precisePressLocation = m; // TODO: use it
        pressBefore = true;
    }

    public void handleRelease(Grid grid, MousePos m) {
        pressBefore = false;
        if (newMode == NewMode.expectingP4) {
            handleNewFinish(grid, m);
        }
        if (pressLocation != null) {
            handleDragOrFinish(grid, pressLocation, m);
            pressLocation = null;
        }
        if (rectangleSelection != null) {
            MousePos from = rectangleSelection;
            MousePos to = m;
            rectangleSelection = null;
            selectByRectangle(from, to);
        }
    }

    private void selectByRectangle(MousePos from, MousePos to) {
        if (!getSelectedCurveStream().findFirst().isPresent()) {
            // no curve is selected, do curve selection
            for (Curve curve : glyph.curves) {
                for (Point point : curve.allPoints()) {
                    if (pointInsideMouseRectangle(from, to, point)) {
                        curve.select();
                    }
                }
            }
            getChangesToCurves().postEvent();
        } else {
            // some curves are selected, do point selection on them
            for (Curve curve : glyph.curves) {
                if (curve.selected) {
                    for (Point point : curve.allPoints()) {
                        if (pointInsideMouseRectangle(from, to, point)) {
                            point.selected = true;
                        }
                    }
                }
            }
            getChangesToCurves().postEvent();
        }
    }

    private static boolean pointInsideMouseRectangle(MousePos from, MousePos to, Point point) {
        return point.x <= Math.max(from.x, to.x)
                && point.x >= Math.min(from.x, to.x)
                && point.y <= Math.max(from.y, to.y)
                && point.y >= Math.min(from.y, to.y);
    }

    static class PointAndDistance implements Comparable<PointAndDistance> {
        Point p;
        double distance;

        public PointAndDistance(Point p, double distance) {
            this.p = p;
            this.distance = distance;
        }

        @Override
        public int compareTo(PointAndDistance o) {
            return Double.compare(distance, o.distance);
        }
    }

    PointAndDistance getPointDistance(Point p, MousePos mp) {
        return new PointAndDistance(p, Math.sqrt(Math.pow(p.x - mp.x, 2) + Math.pow(p.y - mp.y, 2)));
    }

    public boolean handleClick(Grid grid, MousePos m) {
        if (m == null) {
            return false;
        }
        List<PointAndDistance> closest = getClosestPointsAndDistance(m);
        if (m.mod1) {
            return false; // only used for curve selections
        } else if (m.mod2) {
            if (!closest.isEmpty()) {
                boolean change = !closest.get(0).p.selected;
                for (PointAndDistance pd : closest) {
                    pd.p.selected = change;
                    changesToCurves.postEvent();
                }
            }
            return true; // handle even if not within distance to avoid unselecting all
        } else {
            if (!closest.isEmpty()) {
                int selectedClosestCount = 0;
                int firstSelectedClosestIndex = -1;
                int index = 0;
                for (PointAndDistance pd : closest) {
                    if (pd.p.selected) {
                        selectedClosestCount++;
                        if (firstSelectedClosestIndex == -1) {
                            firstSelectedClosestIndex = index;
                        }
                    }
                    index++;
                }
                int selectedCount = (int) getSelectedPointStream().filter(p -> p.selected).count();
                System.out.println("selectedCount: " + selectedCount);
                getSelectedPointStream().forEach(p -> p.selected = false);
                // select next (to select all, use mod2)
                closest.get((firstSelectedClosestIndex + 1) % closest.size()).p.selected = true;
                changesToCurves.postEvent();
                return true;
            }
        }
        return false;
    }

    private List<PointAndDistance> getClosestPointsAndDistance(MousePos m) {
        return getAllPointsInSelectedCurvesStream().map(p -> getPointDistance(p, m)).filter(pd -> pd.distance < m.getClickSize()).sorted().collect(Collectors.toList());
    }

    public void handleCancel() {
        // todo
//        for (CurveDesignerModel curveDesignerModel : getSelectedModels()) {
//            curveDesignerModel.handleCancel();
//        }
//        changesToCurves.postEvent();
    }

    public Stream<Curve> getSelectedCurveStream() {
        return glyph.curves.stream().filter(c -> c.selected);
    }

    public static boolean closePointsRelation(Point p1, Point p2) {
        return almostEquals(p1.x, p2.x) && almostEquals(p1.y, p2.y);
    }

    private static boolean almostEquals(double a, double b) {
        return Math.abs(a - b) < 0.002;
    }


    public Stream<Point> getSelectedPointStream() {
        IdentityHashMap<Point, Point> seen = new IdentityHashMap<Point, Point>();
        List<Point> result = new ArrayList<>();
        getSelectedCurveStream().forEachOrdered(c -> {
            for (Point p : c.allPoints()) {
                if (p.selected) {
                    if (!seen.containsKey(p)) {
                        seen.put(p, p);
                        result.add(p);
                    }
                    for (Point pAssoc : c.allAssociatedPoints(p, GlyphDesignerModel::closePointsRelation)) {
                        if (!seen.containsKey(pAssoc)) {
                            seen.put(pAssoc, pAssoc);
                            result.add(pAssoc);
                        }
                    }
                }
            }
        });
        return result.stream();
    }

    public Stream<Point> getSelectedPointOnlyStream() {
        return getSelectedCurveStream().flatMap(c -> c.allPoints().stream().filter(p -> p.selected));
    }

    public Stream<Point> getSelectedEndPointsStream() {
        return getSelectedCurveStream().flatMap(c -> c.allEndPoints().stream()).filter(p -> p.selected);
    }

    public void moveRel(double dx, double dy) {
        getSelectedCurveStream().forEach(c -> c.moveRel(dx, dy));
        changesToCurves.postEvent();
    }

    private MousePos getPivot() {
        List<Point> points = getSelectedPointStream().limit(2).collect(Collectors.toList());
        if (points.size() == 1) { // exactly one point selected
            Point p = points.get(0);
            return new MousePos(p.x, p.y);
        } else {
            // compute average of selected curve p1 and p4 points
            double[] sums = new double[2];
            int[] count = new int[1];
            getSelectedCurveStream().flatMap(c -> c.allEndPoints().stream()).filter(p -> p.selected).forEach(p -> {
                sums[0] += p.x;
                sums[1] += p.y;
                count[0]++;
            });
            if (count[0] == 0) {
                return null;
            } else {
                return new MousePos(sums[0] / count[0], sums[1] / count[0]);
            }
        }
    }

    public void rotateCCW() {
        MousePos pivot = getPivot();
        if (pivot != null) {
            getSelectedCurveStream().forEach(c -> c.rotateCCW(pivot));
            changesToCurves.postEvent();
        }
    }

    public void rotateCW() {
        MousePos pivot = getPivot();
        if (pivot != null) {
            getSelectedCurveStream().forEach(c -> c.rotateCW(pivot));
            changesToCurves.postEvent();
        }
    }

    public void flipLeftRight() {
        MousePos pivot = getPivot();
        if (pivot != null) {
            getSelectedCurveStream().forEach(c -> c.flipLeftRight(pivot));
            changesToCurves.postEvent();
        }
    }

    public void flipTopBottom() {
        MousePos pivot = getPivot();
        if (pivot != null) {
            getSelectedCurveStream().forEach(c -> c.flipTopBottom(pivot));
            changesToCurves.postEvent();
        }
    }

    public GlyphDefaults getGlyphDefaults() {
        return glyphDefaults;
    }


    public class Removed {
        private final int index;

        public Removed(int index) {
            this.index = index;
        }

        public int getIndex() {
            return index;
        }
    }

    public class MultiRemoved {
    }

    public class Added {
        private final int index;

        public Added(int index) {
            this.index = index;
        }

        public int getIndex() {
            return index;
        }
    }

    @XmlRootElement
    public static class CurveList {
        public ArrayList<Curve> curves = new ArrayList<>();

        public CurveList() {
        }

        public CurveList(List<Curve> curves) {
            this.curves = new ArrayList<>(curves);
        }

        @Override
        public String toString() {
            return "CurveList{" +
                    "curves=" + curves +
                    '}';
        }
    }

    public String serializeSelection() {
        return Persistence.serializeObject(new CurveList(getSelectedCurveStream().collect(Collectors.toList())));
    }

    public String serializeSelectionAsNumber() {
        BigInteger k = BigInteger.valueOf(0);
        int p = 0;
        int g = 1000;
        BigInteger bg = BigInteger.valueOf(g);
        List<Curve> curves = getSelectedCurveStream().collect(Collectors.toList());
        for (Curve curve : curves) {
            k = addCoordinate(k, g, bg, curve.p4.y);
            k = addCoordinate(k, g, bg, curve.p4.x);
            k = addCoordinate(k, g, bg, curve.p3.y);
            k = addCoordinate(k, g, bg, curve.p3.x);
            k = addCoordinate(k, g, bg, curve.p2.y);
            k = addCoordinate(k, g, bg, curve.p2.x);
            k = addCoordinate(k, g, bg, curve.p1.y);
            k = addCoordinate(k, g, bg, curve.p1.x);
            p++;
        }
        String ls = System.lineSeparator();
        return "k=" + k + ls + "p=" + p + ls + "g=" + g;
    }

    private BigInteger addCoordinate(BigInteger k, int g, BigInteger bg, double coordinate) {
        if (coordinate < 0 || coordinate >= 1) {
            throw new RuntimeException("coordinate outside of range 0 to 1: " + coordinate);
        }
        k = k.multiply(bg).add(BigInteger.valueOf((long) (coordinate * g)));
        return k;
    }

    public int pasteSerialized(String data) {
        CurveList curveList = (CurveList) Persistence.deserializeObject(data);
        glyph.curves.addAll(curveList.curves);

        changesToLayout.postEvent();
        changesToCurves.postEvent();

        return curveList.curves.size();
    }

    public CurveList exportData() {
        return new CurveList(glyph.curves);
    }

    public void importData(CurveList dto) {
        glyph.curves.clear();
        glyph.curves.addAll(dto.curves);

        changesToLayout.postEvent();
        changesToCurves.postEvent();
    }


}
