/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.function.Consumer;
import java.util.stream.IntStream;

/**
 */
public class GlyphPainter {
    public static final int VERTEX_SIZE = 15;
    private final ViewZoom viewZoom;
    private BufferedImage bi;
    private boolean forceRedraw;
    private GlyphPainterCache cache;

    public GlyphPainterCache getCache(int w, int h) {
        if (cache == null || cache.w != w || cache.h != h) {
            cache = new GlyphPainterCache(w, h);
        }
        return cache;
    }

    public void paintSelectionRectangle(Dimension size, Graphics g, double x1, double y1, double x2, double y2) {
        CoordinateMapping coordinateMapping = new CoordinateMapping(size, viewZoom);
        int inx1 = coordinateMapping.getInx1();
        int iny1 = coordinateMapping.getIny1();
        int inx2 = coordinateMapping.getInx2();
        int iny2 = coordinateMapping.getIny2();
        int w = inx2 - inx1;
        int h = iny2 - iny1;
        g.setColor(new Color(220, 200, 0));
        int xx1 = transformX(w, x1, inx1, viewZoom);
        int yy1 = transformY(h, y1, iny1, viewZoom);
        int xx2 = transformX(w, x2, inx1, viewZoom);
        int yy2 = transformY(h, y2, iny1, viewZoom);
        int minx = Math.min(xx1, xx2);
        int maxx = Math.max(xx1, xx2);
        int miny = Math.min(yy1, yy2);
        int maxy = Math.max(yy1, yy2);
        g.drawRect(minx, miny, maxx - minx, maxy - miny);
    }

    public static class GlyphPainterCache {
        private final int w, h;
        private long[] selectionHint, selectionHintNext;
        private int selectionAndRenderHintWidth;
        private HashMap<Curve, Long> selectionHintAsCurveCache;
        private HashMap<Curve, Long> selectionHintAsCurveCacheNext;
        private HashMap<Curve, double[]> renderHintAsCurveCache;
        private HashMap<Curve, double[]> renderHintAsCurveCacheNext;

        public GlyphPainterCache(int w, int h) {
            this.w = w;
            this.h = h;
            selectionHint = new long[w * h];
            selectionHintAsCurveCache = new HashMap<>();
            renderHintAsCurveCache = new HashMap<>();
            selectionHintNext = new long[w * h];
            selectionHintAsCurveCacheNext = new HashMap<>();
            renderHintAsCurveCacheNext = new HashMap<>();
            selectionAndRenderHintWidth = w;
        }

        public void flip() {
            HashMap<Curve, Long> tmpCache = selectionHintAsCurveCache;
            selectionHintAsCurveCache = selectionHintAsCurveCacheNext;
            selectionHintAsCurveCacheNext = tmpCache;
            long[] tmpHint = selectionHint;
            selectionHint = selectionHintNext;
            selectionHintNext = tmpHint;
            HashMap<Curve, double[]> tmpCache2 = renderHintAsCurveCache;
            renderHintAsCurveCache = renderHintAsCurveCacheNext;
            renderHintAsCurveCacheNext = tmpCache2;
        }
    }

    public GlyphPainter(ViewZoom viewZoom) {
        this.viewZoom = viewZoom;
    }

    public void paintOnGraphics(Dimension size, Graphics g, Glyph glyph, Grid grid, Consumer<BufferedImage> backgroundPainter) {
        int h = (int) Math.round(size.getHeight());
        int w = (int) Math.round(size.getWidth());
        System.out.println("g width, height " + w + ", " + h);
        long t1 = System.nanoTime();
        boolean newBufferMandatory = bi == null || bi.getHeight() != h || bi.getWidth() != w;
        if (newBufferMandatory || forceRedraw) {
            if (newBufferMandatory) {
                bi = new BufferedImage(w, h, BufferedImage.TYPE_3BYTE_BGR);
                cache = null;
            }
            if (cache == null) {
                cache = new GlyphPainterCache(w, h);
            }
            cleanRaster(size, h, w);
            drawGlyph(size, bi, glyph, grid, viewZoom, cache, backgroundPainter, true, true);
        }
        g.drawImage(bi, 0, 0, null);
        forceRedraw = false;
        long t2 = System.nanoTime();
        System.out.println("paintOnGraphics took " + ((t2 - t1) / 1000000) + " ms");
    }

    private void cleanRaster(Dimension size, int h, int w) {
        CoordinateMapping coordinateMapping = new CoordinateMapping(size, viewZoom);
        int inx2 = coordinateMapping.getInx2();
        int inx1 = coordinateMapping.getInx1();
        int iny2 = coordinateMapping.getIny2();
        int iny1 = coordinateMapping.getIny1();
        WritableRaster r = bi.getRaster();
        int[] iArrayBorder = new int[3];
        iArrayBorder[0] = 240;
        iArrayBorder[1] = 240;
        iArrayBorder[2] = 240;
        int[] iArrayInner = new int[3];
        iArrayInner[0] = 255;
        iArrayInner[1] = 255;
        iArrayInner[2] = 255;
        for (int y = 0; y < h; y++) {
            for (int x = 0; x < w; x++) {
                int[] iArray = (x < inx2 && x >= inx1 && y < iny2 && y >= iny1) ? iArrayInner : iArrayBorder;
                r.setPixel(x, y, iArray);
            }
        }
    }

    public void forceRedraw() {
        forceRedraw = true;
    }

    public void invalidateCache() {
        forceRedraw = true;
        cache = null;
    }

    public static void drawGlyph(Dimension size, BufferedImage bi, Glyph glyph, Grid grid, ViewZoom viewZoom, GlyphPainterCache cache, Consumer<BufferedImage> backgroundPainter, boolean drawSelection, boolean drawGrid) {
        CoordinateMapping coordinateMapping = new CoordinateMapping(size, viewZoom);
        int inx2 = coordinateMapping.getInx2();
        int inx1 = coordinateMapping.getInx1();
        int iny2 = coordinateMapping.getIny2();
        int iny1 = coordinateMapping.getIny1();
        WritableRaster r = bi.getRaster();

        if (backgroundPainter != null) {
            backgroundPainter.accept(bi);
        }

        Arrays.fill(cache.selectionHintNext, 0);
        cache.selectionHintAsCurveCacheNext.clear();
        cache.renderHintAsCurveCacheNext.clear();

        ArrayList<Curve> curves = glyph.curves;
        int[] color = new int[3];
        color[0] = 10;
        color[1] = 10;
        color[2] = 10;
        drawInnerToCache(curves, inx1, inx2, iny1, iny2, cache, viewZoom);
        drawInnerFromCache(color, inx1, inx2, iny1, iny2, cache, r);
        if (drawSelection) {
            for (int i = 0; i < curves.size(); i++) {
                Curve curve = curves.get(i);
                long curveBitMarker = 1L << i;
                drawCurveSkeletonToSelectionCache(curveBitMarker, curve, inx1, inx2, iny1, iny2, bi, cache, viewZoom);
            }
            for (Curve curve : curves) {
                if (curve.selected) {
                    drawCurveSkeletonAnchors(curve, inx1, inx2, iny1, iny2, bi, viewZoom);
                }
            }
            for (int sel = 0; sel <= 1; sel++) {
                for (Curve c : curves) {
                    if (c.selected) {
                        boolean filter = sel == 1;
                        drawPoint(c.p1, PointType.end, filter, inx1, inx2, iny1, iny2, bi, viewZoom);
                        drawPoint(c.p2, PointType.curve, filter, inx1, inx2, iny1, iny2, bi, viewZoom);
                        drawPoint(c.p3, PointType.curve, filter, inx1, inx2, iny1, iny2, bi, viewZoom);
                        drawPoint(c.p4, PointType.end, filter, inx1, inx2, iny1, iny2, bi, viewZoom);
                    }
                }
            }
        }
        if (drawGrid) {
            List<Double> xGuides = grid.getxGuides();
            List<Double> yGuides = grid.getyGuides();

            Integer xdiv = grid.getxGridDivision();
            Integer ydiv = grid.getyGridDivision();
            boolean guidesActive = grid.isGuidesActive();
            if (xdiv != null || ydiv != null || ((!xGuides.isEmpty() || !yGuides.isEmpty()) && guidesActive)) {
                int w1 = inx2 - inx1;
                int h1 = iny2 - iny1;
                double epsX = (0.5 / (double) (w1)) * (viewZoom.getX2() - viewZoom.getX1());
                double epsY = (0.5 / (double) (h1)) * (viewZoom.getY2() - viewZoom.getY1());
                int[] iArray = new int[3];
                for (int y = iny1; y < iny2; y++) {
                    boolean onYGuide = false;
                    boolean onYAxis = false;
                    double yinv = transformYinv(h1, y, iny1, viewZoom);
                    if (ydiv != null) {
                        double yinvdiv = (yinv + epsY / (ydiv + 1)) * ydiv;
                        if (Math.abs(Math.round(yinvdiv) - yinvdiv) <= epsY * ydiv) {
                            onYAxis = true;
                        }
                    }
                    if (guidesActive && Collections.binarySearch(yGuides, yinv - epsY) != Collections.binarySearch(yGuides, yinv + epsY)) {
                        onYGuide = true;
                    }
                    for (int x = inx1; x < inx2; x++) {
                        boolean onAxis = onYAxis;
                        boolean onGuide = onYGuide;
                        double xinv = transformXinv(w1, x, inx1, viewZoom);
                        if (!onYAxis && xdiv != null) {
                            double xinvdiv = (xinv + epsX / (xdiv + 1)) * xdiv;
                            if (Math.abs(Math.round(xinvdiv) - xinvdiv) <= epsX * xdiv) {
                                onAxis = true;
                            }
                        }
                        if (!onYGuide) {
                            if (guidesActive && Collections.binarySearch(xGuides, xinv - epsX) != Collections.binarySearch(xGuides, xinv + epsX)) {
                                onGuide = true;
                            }
                        }
                        if (onGuide) {
                            r.getPixel(x, y, iArray);
                            iArray[0] = Integer.max(0, iArray[0] - 180);
                            iArray[1] = Integer.min(255, iArray[1] + 140);
                            iArray[2] = Integer.max(0, iArray[2] - 180);
                            r.setPixel(x, y, iArray);
                        } else if (onAxis) {
                            r.getPixel(x, y, iArray);
                            iArray[0] = iArray[0] ^ 31;
                            iArray[1] = iArray[1] ^ 31;
                            iArray[2] = iArray[2] ^ 31;
                            r.setPixel(x, y, iArray);
                        }
                    }
                }
            }
        }
        cache.flip();
    }

    private static void drawInnerToCache(ArrayList<Curve> curves, int inx1, int inx2, int iny1, int iny2, GlyphPainterCache cache, ViewZoom viewZoom) {
        long t1 = System.nanoTime();
        int w = inx2 - inx1;
        int h = iny2 - iny1;
        for (Curve curve : curves) {
            double[] doubles = cache.renderHintAsCurveCache.get(curve);
            if (doubles != null) {
                cache.renderHintAsCurveCacheNext.put(curve, doubles);
            } else {
                double[] angles = new double[w * h];
                cache.renderHintAsCurveCacheNext.put(curve, angles);
                IntStream.range(0, h).parallel().forEach(y -> {
                    double yy = transformYinv(h, y, 0, viewZoom);
                    for (int x = 0; x < w; x++) {
                        double xx = transformXinv(w, x, 0, viewZoom);
                        angles[w * y + x] = computeWa(curve, xx, yy);
                    }
                });
            }
        }
        long t2 = System.nanoTime();
        System.out.println("drawInnerToCache took " + ((t2 - t1) / 1000000) + " ms");
    }

    private static void drawInnerFromCache(int[] color, int inx1, int inx2, int iny1, int iny2, GlyphPainterCache cache, WritableRaster r) {
        long t1 = System.nanoTime();
        int w = inx2 - inx1;
        int h = iny2 - iny1;
        double[] angles = new double[w * h];
        Collection<double[]> values = cache.renderHintAsCurveCacheNext.values();
        IntStream.range(0, w * h).parallel().forEach(i -> {
            double sum = 0;
            for (double[] value : values) {
                sum += value[i];
            }
            angles[i] = sum;
        });
        for (int y = 0; y < h; y++) {
            for (int x = 0; x < w; x++) {
//                double wn = angles[w * y + x] / (2 * Math.PI);
//                if ((Math.round(wn) & 1) == 1) {
//                    r.setPixel(x + inx1, y + iny1, color);
//                }
                double wn = angles[w * y + x];
                if (Math.abs(wn) > Math.PI) {
                    r.setPixel(x + inx1, y + iny1, color);
                }
            }
        }
        long t2 = System.nanoTime();
        System.out.println("drawInnerFromCache took " + ((t2 - t1) / 1000000) + " ms");
    }

    private static double computeWa(Curve curve, double xx, double yy) {
        return computeWaSingle(xx, yy, curve.p1.x + epsilon, curve.p1.y + epsilon, curve.p2.x, curve.p2.y, curve.p3.x, curve.p3.y, curve.p4.x + epsilon, curve.p4.y + epsilon);
    }

    private static double computeWaSingle(double xx, double yy, double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4) {
        double boxminx = Double.min(Double.min(x1, x2), Double.min(x3, x4));
        double boxmaxx = Double.max(Double.max(x1, x2), Double.max(x3, x4));
        double boxminy = Double.min(Double.min(y1, y2), Double.min(y3, y4));
        double boxmaxy = Double.max(Double.max(y1, y2), Double.max(y3, y4));
        if (xx < boxminx || xx > boxmaxx || yy < boxminy || yy > boxmaxy) {
            // compute by endpoints
//            double a1 = Math.atan2(y1 - yy, x1 - xx);
//            double a2 = Math.atan2(y4 - yy, x4 - xx);
//            double a = a2 - a1;
            double dy1 = y1 - yy;
            double dx1 = x1 - xx;
            double dy4 = y4 - yy;
            double dx4 = x4 - xx;
            double a = Math.atan2(dy4 * dx1 - dy1 * dx4, dx1 * dx4 + dy1 * dy4);
            return a;
        } else {
            double boxw = boxmaxx - boxminx;
            double boxh = boxmaxy - boxminy;
            double diagonal = Math.sqrt(boxw * boxw + boxh * boxh);
            if (diagonal > 0.001) {// epsilon) {
                // split
                double split1p1x = x1;
                double split1p1y = y1;
                double split2p4x = x4;
                double split2p4y = y4;
                double split1p2x = mid(x1, x2);
                double split1p2y = mid(y1, y2);
                double split2p3x = mid(x3, x4);
                double split2p3y = mid(y3, y4);
                double tmpx = mid(x2, x3);
                double tmpy = mid(y2, y3);
                double split1p3x = mid(tmpx, split1p2x);
                double split1p3y = mid(tmpy, split1p2y);
                double split2p2x = mid(tmpx, split2p3x);
                double split2p2y = mid(tmpy, split2p3y);
                double split1p4_2p1x = mid(split1p3x, split2p2x);
                double split1p4_2p1y = mid(split1p3y, split2p2y);
                return computeWaSingle(xx, yy, split1p1x, split1p1y, split1p2x, split1p2y, split1p3x, split1p3y, split1p4_2p1x, split1p4_2p1y)
                        + computeWaSingle(xx, yy, split1p4_2p1x, split1p4_2p1y, split2p2x, split2p2y, split2p3x, split2p3y, split2p4x, split2p4y);
            } else {
                // about on curve
                return Math.PI;
                // an option to have more smooth curve, to approximate curve using line formed by x1,y1 and x4,y4:
                // return (x4 - x1) * (yy - y1) - (y4 - y1) * (xx - x1) > 0 ? Math.PI : -Math.PI;
            }

        }
    }

    private static double mid(double a, double b) {
        return (a + b) / 2;
    }

    public static final double epsilon = 1e-8;


    public static double bezierCubic(double a1, double a2, double a3, double a4, double t) {
        double oneMinusT = 1 - t;
        double oneMinusT2 = oneMinusT * oneMinusT;
        double oneMinusT3 = oneMinusT2 * oneMinusT;
        double t2 = t * t;
        double t3 = t2 * t;
        return a1 * oneMinusT3 + 3 * a2 * oneMinusT2 * t + 3 * a3 * oneMinusT * t2 + a4 * t3;
    }


    private static void drawCurveSkeletonToSelectionCache(long curveBitMarker, Curve c, int inx1, int inx2, int iny1, int iny2, BufferedImage bi, GlyphPainterCache cache, ViewZoom viewZoom) {
        WritableRaster ra = bi.getRaster();
        boolean selected = c.selected;
        int[] color = new int[3];
        color[1] = 40;
        color[2] = 255;
        int w = inx2 - inx1;
        int h = iny2 - iny1;
        Point p1 = c.p1;
        Point p2 = c.p2;
        Point p3 = c.p3;
        Point p4 = c.p4;
        {
            final int tstepsMin = 40;
            double tInc = 1.0 / tstepsMin;
            boolean last = false;
            double eps = 0.000001;
            double eps1 = 1.0 - eps;
            double bxx = transformXinv(w, inx1 + 1, inx1, viewZoom) - transformXinv(w, inx1, inx1, viewZoom);
            double byy = transformYinv(h, iny1 + 1, iny1, viewZoom) - transformYinv(h, iny1, iny1, viewZoom);
            double distLimit = Math.sqrt(bxx * bxx + byy * byy) / 4;
            for (double tt = eps; !last; tt += tInc) {
                double bx = bezierCubic(p1.x, p2.x, p3.x, p4.x, tt);
                double by = bezierCubic(p1.y, p2.y, p3.y, p4.y, tt);
                if (tt <= eps1) {
                    tInc = 1.0 / tstepsMin; // adaptive tt increments starting at 40 samples per curve, increasing if samples are far apart on screen or they are squarish and angle changes too much
                    while (tInc > 0.000005) {
                        double ttPLus = tt + tInc;
                        double bxPlus = bezierCubic(p1.x, p2.x, p3.x, p4.x, ttPLus);
                        double byPlus = bezierCubic(p1.y, p2.y, p3.y, p4.y, ttPLus);
                        double bxDel = bx - bxPlus;
                        double byDel = by - byPlus;

                        double dist = Math.sqrt(bxDel * bxDel + byDel * byDel);
                        boolean shortEnoughDistance = dist < distLimit;
                        if (shortEnoughDistance) {
                            break;
                        } else {
                            tInc /= 2;
                        }
                    }
                } else {
                    tt = eps1;
                    bx = bezierCubic(p1.x, p2.x, p3.x, p4.x, tt);
                    by = bezierCubic(p1.y, p2.y, p3.y, p4.y, tt);
                    last = true;
                }
                int xx = transformX(w, bx, inx1, viewZoom);
                int yy = transformY(h, by, iny1, viewZoom);
                int r = Math.max(1, VERTEX_SIZE / 3);

                for (int yr = -r; yr < r; yr++) {
                    for (int xr = -r; xr < r; xr++) {
                        if (xr * xr + yr * yr < r * r) {
                            int y = yy + yr;
                            int x = xx + xr;
                            if (x >= inx1 && x < inx2 && y >= iny1 && y < iny2) {
                                cache.selectionHintNext[cache.selectionAndRenderHintWidth * y + x] |= curveBitMarker;
                                if (selected) {
                                    color[0] = (int) Math.round(tt * 255);
                                    ra.setPixel(x, y, color);
                                }
                            }
                        }
                    }
                }
            }
        }
        cache.selectionHintAsCurveCacheNext.put(c.duplicate(), curveBitMarker);
    }

    private static void drawCurveSkeletonAnchors(Curve c, int inx1, int inx2, int iny1, int iny2, BufferedImage bi, ViewZoom viewZoom) {
        int w = inx2 - inx1;
        int h = iny2 - iny1;
        Graphics2D g = bi.createGraphics();
        g.setColor(new Color(0, 160, 255));
        g.setStroke(new BasicStroke(2.0f));
        int ix1 = transformX(w, c.p1.x, inx1, viewZoom);
        int iy1 = transformY(h, c.p1.y, iny1, viewZoom);
        int ix2 = transformX(w, c.p2.x, inx1, viewZoom);
        int iy2 = transformY(h, c.p2.y, iny1, viewZoom);
        int ix3 = transformX(w, c.p3.x, inx1, viewZoom);
        int iy3 = transformY(h, c.p3.y, iny1, viewZoom);
        int ix4 = transformX(w, c.p4.x, inx1, viewZoom);
        int iy4 = transformY(h, c.p4.y, iny1, viewZoom);
        g.setColor(Color.pink);
        g.drawLine(ix1, iy1, ix2, iy2);
        g.drawLine(ix2, iy2, ix3, iy3);
        g.drawLine(ix3, iy3, ix4, iy4);
        g.setColor(Color.red);
        g.drawLine(ix4, iy4, ix1, iy1);
    }

    private static void drawPoint(Point p, PointType pointType, boolean selectionFilter, int inx1, int inx2, int iny1, int iny2, BufferedImage bi, ViewZoom viewZoom) {
        if (p.selected != selectionFilter) {
            return;
        }
        int w = inx2 - inx1;
        int h = iny2 - iny1;
        Graphics2D g = bi.createGraphics();
        int ix = transformX(w, p.x, inx1, viewZoom);
        int iy = transformY(h, p.y, iny1, viewZoom);
        g.setColor(p.selected ? Color.magenta : Color.lightGray);
        g.setStroke(new BasicStroke(2.0f));
        int vs2 = VERTEX_SIZE / 2;
        if (pointType == PointType.end) { // control points 1 and 2
            g.drawRect(ix - vs2, iy - vs2, VERTEX_SIZE, VERTEX_SIZE);
        } else if (pointType == PointType.curve) { // control points 2 and 3
            g.drawOval(ix - vs2, iy - vs2, VERTEX_SIZE, VERTEX_SIZE);
        } else { // control point for cutoff
            int vs4 = VERTEX_SIZE / 4;
            g.drawLine(ix - vs2, iy - vs2, ix - vs4, iy - vs4);
            g.drawLine(ix + vs2, iy - vs2, ix + vs4, iy - vs4);
            g.drawLine(ix - vs2, iy + vs2, ix - vs4, iy + vs4);
            g.drawLine(ix + vs2, iy + vs2, ix + vs4, iy + vs4);
        }
    }

    public ArrayList<Integer> getSelectionHintFor(int x, int y) {
        if (cache != null && cache.selectionHint != null) {
            long hint = cache.selectionHint[cache.selectionAndRenderHintWidth * y + x];
            ArrayList<Integer> curveIndices = new ArrayList<>();
            long j = 1;
            for (int i = 0; i < 64; i++) {
                if ((hint & j) != 0) {
                    curveIndices.add(i);
                }
                j <<= 1;
            }
            return curveIndices;
        } else {
            return new ArrayList<>();
        }
    }

    private static int transformX(int w, double x1, int inx1, ViewZoom viewZoom) {
        x1 = (x1 - viewZoom.getX1()) / (viewZoom.getX2() - viewZoom.getX1());
        return inx1 + (int) (x1 * w);
    }

    private static int transformY(int h, double y1, int iny1, ViewZoom viewZoom) {
        y1 = (y1 - viewZoom.getY1()) / (viewZoom.getY2() - viewZoom.getY1());
        return iny1 + h - 1 - ((int) (y1 * h));
    }

    private static double transformXinv(int w, int x1, int inx1, ViewZoom viewZoom) {
        double zeroToOne = ((double) (x1 - inx1)) / w;
        return viewZoom.getX1() + zeroToOne * (viewZoom.getX2() - viewZoom.getX1());
    }

    private static double transformYinv(int h, int y1, int iny1, ViewZoom viewZoom) {
        double zeroToOne = ((double) (iny1 + h - 1 - y1)) / h;
        return viewZoom.getY1() + zeroToOne * (viewZoom.getY2() - viewZoom.getY1());
    }

}
