/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import javax.xml.bind.annotation.XmlRootElement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 */
public class Grid {
    private int[] xGridDivisions;
    private int[] yGridDivisions;
    private int activeGrid;
    private String xGuidesString;
    private String yGuidesString;
    private List<Double> xGuides;
    private List<Double> yGuides;
    private boolean guidesActive = true;

    public Grid() {
        clear();
    }

    public void clear() {
        xGridDivisions = new int[]{20, 40, 200, 1000};
        yGridDivisions = new int[]{20, 40, 200, 1000};
        activeGrid = -1;
        xGuidesString = "";
        yGuidesString = "";
        xGuides = new ArrayList<>();
        yGuides = new ArrayList<>();
        importEventQueue.postEvent();
    }

    private EventQueue updateEventQueue = new EventQueue();
    private EventQueue importEventQueue = new EventQueue();

    public EventQueue getUpdateEventQueue() {
        return updateEventQueue;
    }

    public EventQueue getImportEventQueue() {
        return importEventQueue;
    }

    public int getActiveGrid() {
        return activeGrid;
    }

    public void setActiveGrid(int activeGrid) {
        if (activeGrid >= -1 && activeGrid < xGridDivisions.length) {
            this.activeGrid = activeGrid;
            updateEventQueue.postEvent();
        } else {
            throw new RuntimeException("activeGrid out of range");
        }
    }

    public Integer getxGridDivision() {
        return activeGrid < 0 ? null : xGridDivisions[activeGrid];
    }

    public void setxGridDivision(Integer xGridDivisions) {
        if (activeGrid >= 0) {
            this.xGridDivisions[activeGrid] = xGridDivisions;
            updateEventQueue.postEvent();
        }
    }

    public Integer getyGridDivision() {
        return activeGrid < 0 ? null : yGridDivisions[activeGrid];
    }

    public void setyGridDivision(Integer yGridDivisions) {
        if (activeGrid >= 0) {
            this.yGridDivisions[activeGrid] = yGridDivisions;
            updateEventQueue.postEvent();
        }
    }

    public int[] getxGridDivisions() {
        return xGridDivisions;
    }

    public int[] getyGridDivisions() {
        return yGridDivisions;
    }

    public List<Double> getxGuides() {
        return xGuides;
    }

    public List<Double> getyGuides() {
        return yGuides;
    }

    public String getxGuidesString() {
        return xGuidesString;
    }

    public void setxGuidesString(String xGuidesString) {
        this.xGuidesString = xGuidesString;
        this.xGuides = convert(xGuidesString);
        updateEventQueue.postEvent();
    }

    public String getyGuidesString() {
        return yGuidesString;
    }

    public void setyGuidesString(String yGuidesString) {
        this.yGuidesString = yGuidesString;
        this.yGuides = convert(yGuidesString);
        updateEventQueue.postEvent();
    }


    public static final Pattern DECIMAL = Pattern.compile("([0-9]+[.][0-9]+)");
    public static final Pattern FRACTIONAL = Pattern.compile("([0-9]+)/([0-9]+)");

    private List<Double> convert(String guidesString) {
        ArrayList<Double> guides = new ArrayList<>();
        String[] split = guidesString.split(" +");
        for (String s : split) {
            if (!s.isEmpty()) {
                if (DECIMAL.matcher(s).matches()) {
                    guides.add(Double.parseDouble(s));
                } else {
                    Matcher m = FRACTIONAL.matcher(s);
                    if (m.matches()) {
                        guides.add(((double) Integer.parseInt(m.group(1))) / Integer.parseInt(m.group(2)));
                    } else {
                        throw new RuntimeException("Could not parse guides");
                    }
                }
            }
        }
        Collections.sort(guides);
        return guides;
    }

    public MousePos clipToGrid(MousePos m) {
        if (m == null) {
            return null;
        }
        return new MousePos(clipX(m.getX()), clipY(m.getY()), m.getClickSize(), m.mod1, m.mod2);
    }

    public double clipX(double x) {
        Integer xGridDivisions = getxGridDivision();
        if (xGridDivisions != null) {
            x = (double) Math.round(x * xGridDivisions) / xGridDivisions;
        }
        return x;
    }

    public double clipY(double y) {
        Integer yGridDivisions = getyGridDivision();
        if (yGridDivisions != null) {
            y = (double) Math.round(y * yGridDivisions) / yGridDivisions;
        }
        return y;
    }

    public void setxGridDivisions(int j, int v) {
        xGridDivisions[j] = v;
        getUpdateEventQueue().postEvent();
    }

    public void setyGridDivisions(int j, int v) {
        yGridDivisions[j] = v;
        getUpdateEventQueue().postEvent();
    }

    public void setGuidesActive(boolean guidesActive) {
        this.guidesActive = guidesActive;
        getUpdateEventQueue().postEvent();
    }

    public boolean isGuidesActive() {
        return guidesActive;
    }

    @XmlRootElement
    public static class GridDTO {
        private int[] xGridDivisions = new int[0];
        private int[] yGridDivisions = new int[0];
        private int activeGrid = -1;
        private String xGuidesString = "";
        private String yGuidesString = "";
        boolean guidesActive = true;

        public int[] getxGridDivisions() {
            return xGridDivisions;
        }

        public void setxGridDivisions(int[] xGridDivisions) {
            this.xGridDivisions = xGridDivisions;
        }

        public int[] getyGridDivisions() {
            return yGridDivisions;
        }

        public void setyGridDivisions(int[] yGridDivisions) {
            this.yGridDivisions = yGridDivisions;
        }

        public int getActiveGrid() {
            return activeGrid;
        }

        public void setActiveGrid(int activeGrid) {
            this.activeGrid = activeGrid;
        }

        public String getxGuidesString() {
            return xGuidesString;
        }

        public void setxGuidesString(String xGuidesString) {
            this.xGuidesString = xGuidesString;
        }

        public String getyGuidesString() {
            return yGuidesString;
        }

        public void setyGuidesString(String yGuidesString) {
            this.yGuidesString = yGuidesString;
        }

        public boolean isGuidesActive() {
            return guidesActive;
        }

        public void setGuidesActive(boolean guidesActive) {
            this.guidesActive = guidesActive;
        }
    }

    public GridDTO exportData() {
        GridDTO dto = new GridDTO();
        dto.activeGrid = activeGrid;
        dto.xGridDivisions = xGridDivisions.clone();
        dto.yGridDivisions = yGridDivisions.clone();
        dto.xGuidesString = xGuidesString;
        dto.yGuidesString = yGuidesString;
        dto.guidesActive = guidesActive;
        return dto;
    }

    public void importData(GridDTO dto) {
        xGridDivisions = dto.xGridDivisions.clone();
        yGridDivisions = dto.yGridDivisions.clone();
        activeGrid = dto.activeGrid;
        guidesActive = dto.guidesActive;
        setxGuidesString(dto.xGuidesString);
        setyGuidesString(dto.yGuidesString);
        getImportEventQueue().postEvent();
    }
}
