/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import javax.swing.*;
import java.awt.*;

/**
 */
public class HistoryLister extends JFrame {

    private final JList<StateManager.StateHistory> jList;

    public HistoryLister(StateManager stateManager) throws HeadlessException {
        setBounds(new Rectangle(800, 400));
        setTitle("Undo history");
        setLayout(new BorderLayout());
        AbstractListModel<StateManager.StateHistory> abstractListModel = new AbstractListModel<StateManager.StateHistory>() {
            private int lastSize;

            {
                stateManager.getHistoryChangesEventQueue().addEventConsumer(e -> {
                            int size = stateManager.getStateHistory().size();
                            if (e == StateManager.StateHistoryEventType.added) {
                                fireIntervalAdded(this, lastSize, size);
                            } else if (e == StateManager.StateHistoryEventType.replacedLast) {
                                fireContentsChanged(this, lastSize, size);
                            } else if (e == StateManager.StateHistoryEventType.cleared) {
                                fireContentsChanged(this, 0, Math.max(lastSize, size));
                            } else {
                                throw new RuntimeException("unexpected even type");
                            }
                        }
                );
            }

            @Override
            public int getSize() {
                lastSize = stateManager.getStateHistory().size();
                return lastSize;
            }

            @Override
            public StateManager.StateHistory getElementAt(int index) {
                return stateManager.getStateHistory().get(index);
            }
        };
        jList = new JList<>(abstractListModel);
        ViewZoom viewZoom = new ViewZoom();
        GlyphPainter previewGlyphPainter = new GlyphPainter(viewZoom);
        GlyphDesignerModel previewGlyphDesignerModel = new GlyphDesignerModel();
        Grid previewGrid = new Grid();
        StateManager previewStateManager = new StateManager(previewGlyphDesignerModel, previewGrid);
        JPanel previewRenderPanel = new JPanel() {
            @Override
            protected void paintComponent(Graphics g) {
                super.paintComponent(g);
                previewGlyphPainter.paintOnGraphics(getSize(), g, previewGlyphDesignerModel.glyph, previewGrid, null);
            }
        };
        previewRenderPanel.setMinimumSize(new Dimension(300, 300));
        JButton undoButton = new JButton("Undo to this item");
        undoButton.setEnabled(false);
        JLabel detailLabel = new JLabel();
        jList.addListSelectionListener(e -> {
            if (!e.getValueIsAdjusting()) {
                update(stateManager, jList, detailLabel, previewGlyphPainter, previewStateManager, previewRenderPanel);
            }
            undoButton.setEnabled(jList.getSelectedIndex() >= 0);
        });
        stateManager.getHistoryChangesEventQueue().addEventConsumer(e -> {
            if (e == StateManager.StateHistoryEventType.replacedLast) {
                update(stateManager, jList, detailLabel, previewGlyphPainter, previewStateManager, previewRenderPanel);
            }
        });
        JScrollPane jScrollPane = new JScrollPane();
        jScrollPane.setViewportView(jList);
        add(jScrollPane, BorderLayout.WEST);
        add(previewRenderPanel, BorderLayout.CENTER);
        JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new FlowLayout());
        undoButton.addActionListener(e -> {
            int selectedIndex = jList.getSelectedIndex();
            if (selectedIndex >= 0 && selectedIndex < stateManager.getStateHistory().size()) {
                stateManager.undoToState(stateManager.getStateHistory().get(selectedIndex));
            }
        });
        buttonPanel.add(undoButton);
        buttonPanel.add(detailLabel);
        add(buttonPanel, BorderLayout.SOUTH);
    }

    private void update(StateManager stateManager, JList<StateManager.StateHistory> jList, JLabel detailLabel, GlyphPainter previewGlyphPainter, StateManager previewStateManager, JPanel previewRenderPanel) {
        int selectedIndex = jList.getSelectedIndex();
        if (selectedIndex >= 0 && selectedIndex < stateManager.getStateHistory().size()) {
            StateManager.StateHistory stateHistory = stateManager.getStateHistory().get(selectedIndex);
            detailLabel.setText(stateHistory.toLongString());
            previewStateManager.loadDTO(stateHistory.glyphDesignerDTO, true);
            previewGlyphPainter.forceRedraw();
            previewRenderPanel.repaint();
        } else {
            detailLabel.setText("");
        }
    }

    public void showWindow() {
        jList.setSelectedIndex(jList.getModel().getSize() - 1);
        setVisible(true);
    }
}
