/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer;

import javax.xml.bind.annotation.XmlRootElement;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 */
public class StateManager {

    public static final long TIME_GROUP = 2000; // ms, groups of actions in history
    public static final String FILE_EXT_GLYPH = ".glyph";
    public static final String FILE_EXT_GLYPH_DESCRIPTION = "Glyph Designer files";
    private long lastAddTimestamp;
    private boolean actionInProgress;

    enum StateHistoryEventType {
        cleared, added, replacedLast
    }

    public static class StateHistory {
        public long timestamp;
        public ArrayList<String> labels;
        public GlyphDesignerDTO glyphDesignerDTO;

        public StateHistory(long timestamp, ArrayList<String> labels, GlyphDesignerDTO glyphDesignerDTO) {
            this.timestamp = timestamp;
            this.labels = labels;
            this.glyphDesignerDTO = glyphDesignerDTO;
        }

        @Override
        public String toString() {
            return "<html>" + formatTimestamp(timestamp) + "<br/>" + shortenString(30, String.join(". ", labels)) + "</html>";
        }

        private String shortenString(int count, String str) {
            if (str.length() > count) {
                return str.substring(0, count - 3) + "...";
            } else {
                return str;
            }
        }

        public String toLongString() {
            return "<html>" + formatTimestamp(timestamp) + "<br/>" + String.join(". ", labels) + "</html>";
        }
    }

    private final GlyphDesignerModel glyphDesignerModel;
    private final Grid grid;
    private String directory = ".";
    private String glyphName = "";
    private File lastFile;
    private final List<StateHistory> stateHistory = new ArrayList<>();
    private final List<String> labels = new ArrayList<>();
    private final EventQueue historyChangesEventQueue = new EventQueue();

    public StateManager(GlyphDesignerModel glyphDesignerModel, Grid grid) {
        this.glyphDesignerModel = glyphDesignerModel;
        this.glyphDesignerModel.getChangesToCurves().addEventConsumer(e -> recordState());
        this.glyphDesignerModel.getChangesToLayout().addEventConsumer(e -> recordState());
        this.grid = grid;
    }

    public List<StateHistory> getStateHistory() {
        return stateHistory;
    }

    public EventQueue getHistoryChangesEventQueue() {
        return historyChangesEventQueue;
    }

    private void recordState() {
        ArrayList<String> addLabels = new ArrayList<>(this.labels);
        this.labels.clear();
        StateHistory previousSH = this.stateHistory.isEmpty() ? null : this.stateHistory.get(this.stateHistory.size() - 1);
        GlyphDesignerDTO newDTO = getGlyphDesignerDTO();
        if (previousSH != null && (actionInProgress ||
                (addLabels.isEmpty()
                        && (lastAddTimestamp + TIME_GROUP > System.currentTimeMillis() || myDeepEquals(previousSH.glyphDesignerDTO, newDTO))))) {
            addLabels.addAll(previousSH.labels);
            StateHistory sh = new StateHistory(System.currentTimeMillis(), addLabels, deepCopy(newDTO));
            this.stateHistory.set(this.stateHistory.size() - 1, sh);
            historyChangesEventQueue.postEvent(StateHistoryEventType.replacedLast);
        } else {
            StateHistory sh = new StateHistory(System.currentTimeMillis(), addLabels, deepCopy(newDTO));
            stateHistory.add(sh);
            lastAddTimestamp = sh.timestamp;
            historyChangesEventQueue.postEvent(StateHistoryEventType.added);
        }
    }


    public void undoToState(StateHistory stateHistory) {
        historyChangesEventQueue.mergeEventsDuring(() -> {
            loadDTO(stateHistory.glyphDesignerDTO, false);
            addLabel("undo to history item " + formatTimestamp(stateHistory.timestamp));
            recordState();
        });
    }

    private boolean myDeepEquals(GlyphDesignerDTO glyphDesignerDTO, GlyphDesignerDTO newDTO) {
        return Persistence.serializeObject(glyphDesignerDTO).equals(Persistence.serializeObject(newDTO));
    }


    public void mouseStart() {
        actionInProgress = true;
    }

    public void mouseEnd() {
        actionInProgress = false;
        recordState();
    }

    private GlyphDesignerDTO deepCopy(GlyphDesignerDTO glyphDesignerDTO) {
        return (GlyphDesignerDTO) Persistence.deserializeObject(Persistence.serializeObject(glyphDesignerDTO));
    }

    public void addLabel(String label) {
        labels.add(label);
    }

    public String getDirectory() {
        return directory;
    }

    public void clear(boolean clearHistory) {
        grid.clear();
        glyphDesignerModel.clear();
        if (clearHistory) {
            stateHistory.clear();
        }
        lastFile = null;
        historyChangesEventQueue.postEvent(StateHistoryEventType.cleared);
    }

    @XmlRootElement
    public static class GlyphDesignerDTO {
        GlyphDesignerModel.CurveList curveList = new GlyphDesignerModel.CurveList();
        Grid.GridDTO gridDTO = new Grid.GridDTO();

        public GlyphDesignerModel.CurveList getCurveList() {
            return curveList;
        }

        public void setCurveList(GlyphDesignerModel.CurveList curveList) {
            this.curveList = curveList;
        }

        public Grid.GridDTO getGridDTO() {
            return gridDTO;
        }

        public void setGridDTO(Grid.GridDTO gridDTO) {
            this.gridDTO = gridDTO;
        }
    }

    public void setDirectory(String directory) {
        this.directory = directory;
    }

    public void setGlyphName(String glyphName) {
        this.glyphName = glyphName;
    }

    public File getLastFile() {
        return lastFile;
    }

    public static String formatTimestamp(long timestamp) {
        return new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ").format(new Date(timestamp));
    }

    public void loadAction() {
        File file = getFile();
        try {
            String data = new String(Files.readAllBytes(file.toPath()), StandardCharsets.UTF_8);
            GlyphDesignerDTO glyphDesignerDTO = (GlyphDesignerDTO) Persistence.deserializeObject(data);
            loadDTO(glyphDesignerDTO, true);
            lastFile = file;
            addLabel("loaded from " + file);
            recordState();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public void loadDTO(GlyphDesignerDTO glyphDesignerDTO, boolean clearHistory) {
        clear(clearHistory);
        grid.importData(glyphDesignerDTO.gridDTO);
        glyphDesignerModel.importData(glyphDesignerDTO.curveList);
    }

    public File getFile() {
        return new File(directory, glyphName + ".glyph");
    }

    public boolean isFileEmpty() {
        return glyphName.isEmpty();
    }

    public String getGlyphName() {
        return glyphName;
    }

    public void saveAction() {
        GlyphDesignerDTO glyphDesignerDTO = getGlyphDesignerDTO();
        String data = Persistence.serializeObject(glyphDesignerDTO);
        File file = getFile();
        try {
            Files.write(file.toPath(), data.getBytes(StandardCharsets.UTF_8));
            lastFile = file;
            addLabel("saved to " + file);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private GlyphDesignerDTO getGlyphDesignerDTO() {
        GlyphDesignerModel.CurveList curveList = glyphDesignerModel.exportData();
        Grid.GridDTO gridDTO = grid.exportData();
        GlyphDesignerDTO glyphDesignerDTO = new GlyphDesignerDTO();
        glyphDesignerDTO.curveList = curveList;
        glyphDesignerDTO.gridDTO = gridDTO;
        return glyphDesignerDTO;
    }


}
