/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.designer.numbercompositor;

import cz.jtra.math.quine.designer.Curve;
import cz.jtra.math.quine.designer.GlyphDesignerModel;
import cz.jtra.math.quine.designer.Persistence;
import cz.jtra.math.quine.designer.StateManager;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;

/**
 */
public class NumberCompositor {


    public static void main(String[] args) {
        final String glyphDirectory = args[0];
        final String[] includeGlyphs = new String[]{"0-1", "1-1", "2-1", "3-1", "4-1", "5-1", "6-1", "7-1", "8-1", "9-1", "__n1", "_equal1"};
        final TypesetGlyph[] typesetGlyphs = new TypesetGlyph[]{
                ts(10, 0.8, 0, 5),
                ts(11, 1.27, 0, 5)
        };

        String s = composeNumber(glyphDirectory, includeGlyphs, typesetGlyphs);
        System.out.println("number:");
        System.out.println(s);
    }

    public static String composeNumber(String glyphDirectory, String[] includeGlyphs, TypesetGlyph[] typesetGlyphs) {
        int numGlyphs = includeGlyphs.length;
        System.out.println("numGlyphs = " + numGlyphs);
        int minimalCurvePointStart = (numGlyphs + 1) * 5 + 2;
        System.out.println("curvePointStart = " + minimalCurvePointStart);
        StringBuilder numberIndexRev = new StringBuilder();
        StringBuilder numberGlyphsRev = new StringBuilder();
        int index = minimalCurvePointStart;
        for (String includeGlyph : includeGlyphs) {
            String glyphDigits = loadGlyph(glyphDirectory, includeGlyph);
            numberGlyphsRev.append(new StringBuilder(glyphDigits).reverse());
            numberIndexRev.append(new StringBuilder(glyphIndexFormat(index)).reverse());
            index += glyphDigits.length();
        }
        numberIndexRev.append(new StringBuilder(glyphIndexFormat(index)).reverse());
        System.out.println("typesetDataStart = " + index);
        StringBuilder outputRev = new StringBuilder();
        outputRev.append(new StringBuilder(String.format("%02d", numGlyphs)).reverse());
        outputRev.append(numberIndexRev);
        outputRev.append(numberGlyphsRev);
        for (TypesetGlyph typesetGlyph : typesetGlyphs) {
            outputRev.append(typesetGlyph.toStringBuilder().reverse());
        }
        return outputRev.reverse().toString();
    }

    private static String loadGlyph(String glyphDirectory, String glyphName) {
        File file = new File(glyphDirectory, glyphName + ".glyph");
        StringBuilder sb = new StringBuilder();
        try {
            String data = new String(Files.readAllBytes(file.toPath()), StandardCharsets.UTF_8);
            StateManager.GlyphDesignerDTO glyphDesignerDTO = (StateManager.GlyphDesignerDTO) Persistence.deserializeObject(data);
            GlyphDesignerModel.CurveList curveList = glyphDesignerDTO.getCurveList();
            for (Curve curve : curveList.curves) {
                sb.append(glyphPointFormat(curve.p4.y));
                sb.append(glyphPointFormat(curve.p4.x));
                sb.append(glyphPointFormat(curve.p3.y));
                sb.append(glyphPointFormat(curve.p3.x));
                sb.append(glyphPointFormat(curve.p2.y));
                sb.append(glyphPointFormat(curve.p2.x));
                sb.append(glyphPointFormat(curve.p1.y));
                sb.append(glyphPointFormat(curve.p1.x));
            }
            return sb.toString();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private static void padZerosRight(int n, StringBuilder sb) {
        if (sb.length() > n) {
            throw new RuntimeException("input too large to pad to smaller size");
        }
        for (int i = sb.length(); i < n; i++) {
            sb.append('0');
        }
    }

    private static String glyphIndexFormat(int i) {
        return String.format("%05d", i);
    }

    private static String glyphPointFormat(double d) {
        if (d < 0) {
            throw new RuntimeException("Negative numbers not allowed: " + d);
        }
        if (d >= 1) {
            throw new RuntimeException("Negative numbers equal and above 1 not allowed: " + d);
        }
        double dRound = 0.0005;
        if (d + dRound < 1.0d) {
            d += dRound;
        }
        int i = (int) (d * 1000);
        return String.format("%03d", i);
    }

    private static String glyphPositionFormat(double d) {
        if (d < 0) {
            throw new RuntimeException("Negative numbers not allowed in glyph target position: " + d);
        }
        if (d >= 100) {
            throw new RuntimeException("Positive numbers equal and above 100 not allowed in glyph target position: " + d);
        }
        double dRound = 0.0005;
        if (d + dRound < 100.0d) {
            d += dRound;
        }
        int i = (int) (d * 1000);
        return String.format("%05d", i);
    }

    public static TypesetGlyph ts(int glyphNr, double positionX, double positionY, int scaling) {
        return new TypesetGlyph(glyphNr, positionX, positionY, scaling);
    }

    public static class TypesetGlyph {
        int glyphNr; // 2 digits
        double positionX, positionY; // 5 digits in decimal format XX.XXX
        int scaling; // 1 digit: scale=1.25^(scaling-5)

        public TypesetGlyph(int glyphNr, double positionX, double positionY, int scaling) {
            if (glyphNr >= 0 && glyphNr <= 99) {
                this.glyphNr = glyphNr;
            } else {
                throw new RuntimeException("glyphNr out of range: " + glyphNr);
            }
            this.positionX = positionX;
            this.positionY = positionY;
            if (scaling >= 0 && scaling <= 9) {
                this.scaling = scaling;
            } else {
                throw new RuntimeException("scaling out of range: " + scaling);
            }

        }

        public StringBuilder toStringBuilder() {
            StringBuilder sb = new StringBuilder(13);
            sb.append(scaling);
            sb.append(glyphPositionFormat(positionY));
            sb.append(glyphPositionFormat(positionX));
            sb.append(String.format("%02d", glyphNr));
            return sb;
        }

        @Override
        public String toString() {
            return "TypesetGlyph{" +
                    "glyphNr=" + glyphNr +
                    ", positionX=" + positionX +
                    ", positionY=" + positionY +
                    ", scaling=" + scaling +
                    '}';
        }
    }
}
