/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.renderer;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

public class MergeImages {

    public static final String MERGED_PNG = "merged.png";

    public static void main(String[] args) {
        if (args.length < 2) {
            System.err.println("Expecting at least 2 PNG image file names.");
            System.err.println("Images must have same dimensions. Images must have 4 channels (RGB and alpha).");
            System.err.println("Merging does not use gamma.");
            System.err.println("Merged output will be written to file " + MERGED_PNG);
            System.exit(1);
        }
        long t1 = System.currentTimeMillis();
        List<BufferedImage> images = Arrays.stream(args).map(fn -> {
            System.out.println("loading image " + fn);
            return loadImage(fn);
        }).collect(Collectors.toList());
        int width = images.get(0).getWidth();
        int height = images.get(0).getHeight();
        int imageCount = images.size();
        Raster[] rasters = new Raster[imageCount];
        for (int i = 0; i < imageCount; i++) {
            BufferedImage image = images.get(i);
            if (image.getWidth() != width) {
                System.err.println("Image " + args[i] + " has width " + image.getWidth() + " but expecting " + width);
                System.exit(2);
            }
            if (image.getHeight() != height) {
                System.err.println("Image " + args[i] + " has height " + image.getHeight() + " but expecting " + height);
                System.exit(3);
            }
            rasters[i] = image.getRaster();
        }
        System.out.println("merging");
        BufferedImage bi = new BufferedImage(width, height, BufferedImage.TYPE_4BYTE_ABGR);
        WritableRaster r = bi.getRaster();
        IntStream.range(0, height).parallel().forEach(y -> {
            double[] sums = new double[4];
            double[] pixel = new double[4];
            for (int x = 0; x < width; x++) {
                sums[0] = 0;
                sums[1] = 0;
                sums[2] = 0;
                sums[3] = 0;
                for (int i = 0; i < imageCount; i++) {
                    rasters[i].getPixel(x, y, pixel);
                    sums[0] += pixel[0];
                    sums[1] += pixel[1];
                    sums[2] += pixel[2];
                    sums[3] += pixel[3];
                }
                sums[0] = sums[0] / imageCount;
                sums[1] = sums[1] / imageCount;
                sums[2] = sums[2] / imageCount;
                sums[3] = sums[3] / imageCount;
                r.setPixel(x, y, sums);
            }
        });
        saveImage(bi, MERGED_PNG);
        long t2 = System.currentTimeMillis();
        System.out.println("took: " + (t2 - t1) / 1000.0 + "s");
    }

    private synchronized static void saveImage(BufferedImage bi, String fileName) {
        try {
            File file = new File(fileName);
            ImageIO.write(bi, "png", file);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private synchronized static BufferedImage loadImage(String name) {
        try {
            return ImageIO.read(new File(name));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

}
