/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.renderer;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.io.*;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.IntStream;

public class RenderJob {

    public static final String PREVIOUS_PNG = ".previous.png";

    public static void main(String[] args) throws InterruptedException, IOException, IllegalAccessException, InstantiationException {
        long t1 = System.currentTimeMillis();
        if (args.length != 1) {
            System.err.println("expecting job file that contains lines:");
            System.err.println("1. output image file name (PNG format, also used for input if restarted)");
            System.err.println("2. short class name of drawing class");
            System.err.println("3. saving interval in seconds (use longer for big images)");
            System.err.println("4. width of resulting image");
            System.err.println("5. view port x1");
            System.err.println("6. view port x2");
            System.err.println("7. view port y1");
            System.err.println("8. view port y2");
            System.err.println("9. render x1");
            System.err.println("10. render x2");
            System.err.println("11. render y1");
            System.err.println("12. render y2");
            System.err.println("13. subpixel grid (1=no grid)");
            System.err.println("14. subpixel shift x (0=no shift)");
            System.err.println("15. subpixel shift y (0=no shift)");
            System.err.println("14+ arguments to drawing class");
            System.exit(1);
        }
        File file = new File(args[0]);
        BufferedReader br = new BufferedReader(new FileReader(file));
        String description = br.readLine();
        System.out.println("description: " + description);
        String outputFileName = br.readLine();
        if (!outputFileName.endsWith(".png")) {
            System.err.println("Only .png files are supported, change file name.");
            System.exit(2);
        }
        String className = br.readLine();
        long saveIntervalMs = Long.parseLong(br.readLine()) * 1000L;
        int imageWidth = Integer.parseInt(br.readLine());
        double viewPortX1 = Double.parseDouble(br.readLine());
        double viewPortX2 = Double.parseDouble(br.readLine());
        double viewPortY1 = Double.parseDouble(br.readLine());
        double viewPortY2 = Double.parseDouble(br.readLine());
        double renderX1 = Double.parseDouble(br.readLine());
        double renderX2 = Double.parseDouble(br.readLine());
        double renderY1 = Double.parseDouble(br.readLine());
        double renderY2 = Double.parseDouble(br.readLine());
        int subpixelGrid = Integer.parseInt(br.readLine());
        int subpixelShiftXInt = Integer.parseInt(br.readLine());
        int subpixelShiftYInt = Integer.parseInt(br.readLine());
        double pixelSizeInViewPort = (viewPortX2 - viewPortX1) / imageWidth;
        double subpixelShiftX = (pixelSizeInViewPort / subpixelGrid) * subpixelShiftXInt;
        double subpixelShiftY = (pixelSizeInViewPort / subpixelGrid) * subpixelShiftYInt;
        viewPortX1 += subpixelShiftX;
        viewPortX2 += subpixelShiftX;
        viewPortY1 += subpixelShiftY;
        viewPortY2 += subpixelShiftY;
        renderX1 += subpixelShiftX;
        renderX2 += subpixelShiftX;
        renderY1 += subpixelShiftY;
        renderY2 += subpixelShiftY;
        ArrayList<String> inputs = new ArrayList<>();
        String line;
        while ((line = br.readLine()) != null) {
            inputs.add(line);
        }
        String packageName = RenderJob.class.getPackage().getName();
        Class<?> aClass = null;
        try {
            aClass = Class.forName(packageName + "." + className);
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
            System.exit(3);
        }
        if (!Boolean2dFun.class.isAssignableFrom(aClass)) {
            System.err.println("Specified class should be instance of " + Boolean2dFun.class.getName());
        }
        Boolean2dFun b2df = (Boolean2dFun) aClass.newInstance();
        b2df.init(inputs);
        renderFunction(viewPortX1, viewPortX2, viewPortY1, viewPortY2, renderX1, renderX2, renderY1, renderY2, imageWidth, outputFileName, saveIntervalMs, b2df);
        long t2 = System.currentTimeMillis();
        System.out.println("took: " + (t2 - t1) / 1000.0 + "s");
    }

    public interface Boolean2dFun {
        void init(List<String> args);

        boolean fun(double x, double y);
    }


    private static void renderFunction(double viewPortX1, double viewPortX2, double viewPortY1, double viewPortY2,
                                       double renderX1, double renderX2, double renderY1, double renderY2,
                                       int imageWidth, String fileName, long saveIntervalMs, Boolean2dFun fun) {
        double pixelSizeInViewPort = (viewPortX2 - viewPortX1) / imageWidth;
        int imageHeight = (int) ((viewPortY2 - viewPortY1) / pixelSizeInViewPort);
        System.out.println("image dimensions: w=" + imageWidth + " h=" + imageHeight);
        int pixels = 0;
        BufferedImage bi;
        WritableRaster r;
        File filePrimary = new File(fileName);
        File filePrevious = new File(fileName + PREVIOUS_PNG);
        if (filePrimary.exists() || filePrevious.exists()) {
            BufferedImage biTmp;
            try {
                biTmp = loadImage(fileName);
                biTmp.getWidth(); // NPE if empty file
                System.out.println("saved image loaded");
            } catch (RuntimeException e) {
                System.out.println("primary image load failed, loading previous image");
                // try previous, current might be incomplete or missing
                biTmp = loadImage(fileName + PREVIOUS_PNG);
                biTmp.getWidth(); // NPE if empty file
                System.out.println("saved previous image loaded");
                System.out.println("successfully loaded previous image" +
                        (filePrimary.exists() ? ", deleting primary image file" : ""));
                // success, delete incomplete image
                filePrimary.delete();
            }
            bi = biTmp;
            if (bi.getHeight() != imageHeight || bi.getWidth() != imageWidth) {
                System.err.println("cannot load existing image incompatible dimensions: " +
                        "w=" + bi.getWidth() + " h=" + bi.getHeight() + " " +
                        "expected: w=" + imageWidth + " h=" + imageHeight);
                System.exit(4);
            }
            r = bi.getRaster();
            int[] readPixel = new int[4];
            for (int x = 0; x < imageWidth; x++) {
                for (int y = 0; y < imageHeight; y++) {
                    double xd = viewPortX1 + x * pixelSizeInViewPort;
                    double yd = viewPortY2 - y * pixelSizeInViewPort;
                    if (!(xd < renderX1
                            || xd > renderX2
                            || yd < renderY1
                            || yd > renderY2)) {
                        r.getPixel(x, y, readPixel);
                        if (readPixel[3] == 0) {
                            pixels++;
                        }
                    }
                }
            }
        } else {
            bi = new BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_4BYTE_ABGR);
            r = bi.getRaster();
            int[] transparent = new int[]{0, 0, 0, 0};
            for (int x = 0; x < imageWidth; x++) {
                for (int y = 0; y < imageHeight; y++) {
                    r.setPixel(x, y, transparent);
                    double xd = viewPortX1 + x * pixelSizeInViewPort;
                    double yd = viewPortY2 - y * pixelSizeInViewPort;
                    if (!(xd < renderX1
                            || xd > renderX2
                            || yd < renderY1
                            || yd > renderY2)) {
                        pixels++;
                    }
                }
            }
        }
        int pixelFinal = pixels;
        int[] white = new int[]{255, 255, 255, 255};
        int[] black = new int[]{0, 0, 0, 255};
        final AtomicInteger count = new AtomicInteger();
        final AtomicLong nextSave = new AtomicLong(System.currentTimeMillis() + saveIntervalMs);

        int size = Runtime.getRuntime().availableProcessors();
        ExecutorService service = new ThreadPoolExecutor(size, size, 500L, TimeUnit.MILLISECONDS,
                new ArrayBlockingQueue<>(10 * size), new ThreadPoolExecutor.CallerRunsPolicy());

        // partial render of horizontal lines every 30 pixels
        IntStream.range(0, (imageWidth / 30) * imageHeight).forEach(xy -> {
            int x = xy % imageWidth;
            int yTmp = xy / imageWidth;
            int y = yTmp * 30;
            renderInner(viewPortX1, viewPortY2, renderX1, renderX2, renderY1, renderY2, fileName, fun, pixelSizeInViewPort,
                    bi, r, pixelFinal, white, black, count, nextSave, saveIntervalMs, service, x, y);
        });
        // full render with stripes
        IntStream.range(0, imageWidth * imageHeight).forEach(xy -> {
            int x = xy % imageWidth;
            int yTmp = xy / imageWidth;
            int yStride = (yTmp & 0xffffff00) | ((yTmp >> 4) & 0xf) | ((yTmp << 4) & 0xf0);
            int y = yStride >= imageHeight ? yTmp : yStride;
            renderInner(viewPortX1, viewPortY2, renderX1, renderX2, renderY1, renderY2, fileName, fun, pixelSizeInViewPort,
                    bi, r, pixelFinal, white, black, count, nextSave, saveIntervalMs, service, x, y);
        });
        service.shutdown();
        try {
            while (!service.awaitTermination(60, TimeUnit.SECONDS)) {
                System.out.println("waiting for task completion");
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        saveImage(bi, fileName);
        filePrevious.delete();
    }

    private static void renderInner(double viewPortX1, double viewPortY2, double renderX1, double renderX2, double renderY1, double renderY2,
                                    String fileName, Boolean2dFun fun, double pixelSizeInViewPort, BufferedImage bi, WritableRaster r,
                                    int pixelFinal, int[] white, int[] black, AtomicInteger count, AtomicLong nextSave, long saveIntervalMs,
                                    ExecutorService service, int x, int y) {
        double xd = viewPortX1 + x * pixelSizeInViewPort;
        double yd = viewPortY2 - y * pixelSizeInViewPort;
        if (!(xd < renderX1
                || xd > renderX2
                || yd < renderY1
                || yd > renderY2)) {
            int[] readPixel = r.getPixel(x, y, (int[]) null);
            if (readPixel[3] == 0) {
                service.submit(() -> {
                    int currentCount = count.getAndIncrement();
                    if (((x + y) % 100) == 0) {
                        System.out.println("x=" + x + " y=" + y + " total=" + currentCount + "/" + pixelFinal);
                    }
                    boolean pixel = fun.fun(xd, yd);
                    r.setPixel(x, y, pixel ? black : white);
                    long t2 = System.currentTimeMillis();
                    long nextSaveCurrent = nextSave.get();
                    if (t2 > nextSaveCurrent) {
                        if (nextSave.compareAndSet(nextSaveCurrent, t2 + saveIntervalMs)) {
                            currentCount = count.get();
                            System.out.println("x=" + x + " y=" + y + " total=" + currentCount + "/" + pixelFinal);
                            System.out.println("saving partial render, next save: "
                                    + new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ").format(new Date(nextSave.get())));
                            saveImage(bi, fileName); // intermediate save
                            long t3 = System.currentTimeMillis();
                            System.out.println("saving done, took " + (t3 - t2) / 1000 + "s");
                        }
                    }
                });
            }
        }
    }

    private synchronized static void saveImage(BufferedImage bi, String fileName) {
        try {
            File file = new File(fileName);
            if (file.exists()) {
                File renameTo = new File(fileName + PREVIOUS_PNG);
                renameTo.delete();
                file.renameTo(renameTo);
            }
            ImageIO.write(bi, "png", file);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private synchronized static BufferedImage loadImage(String name) {
        try {
            return ImageIO.read(new File(name));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
}
