/*
 * MIT License
 *
 * Copyright (c) 2019 Jakub Travnik
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package cz.jtra.math.quine.renderer;

import java.math.BigInteger;
import java.util.List;

import static java.lang.Double.max;
import static java.lang.Double.min;

/*

 The same as SelfFormulaFast, but this variant requires to pass epsilon as second parameter.

*/

/*
Plot Sr(x,y) where x _elem _rat, y _elem _rat

Sr(x, y) > Pi
Sr(x, y) = PrintFormula(x, y) + PrintN(x, y, floor(log10(N)), 0, 2, 0)
PrintN(x, y, C, k, x0, y0) = if k>C then 0; else PrintN(x, y, C, k+1, x0, y0) + Glyph(x, y, x0+(k mod 100)*0.4, y0-0.8*floor(k/100), 1, Dig(N,C-k,1))
Glyph(x, y, x0, y0, s, d) = Evse((x-x0)/s, (y-y0)/s, Dig(N, d*5+2, 5), Dig(N, d*5+7, 5))
Evse(x, y, s, e) = Evk(x, y, Dig(N, s, e-s), (e-s)/24)
Evk(x, y, k, p) = Lim eps->0+ |Evs(x, y, k, p, eps)|
Dig(k, p, d) = floor(k/10^p) mod 10^d
L = Dig(N, Dig(N, 0, 2)*5+2, 5)
PrintFormula(x, y) = PrintFrom(x, y, Dig(N, L, 13), L+13)
PrintFrom(x, y, k, p) = 0, if k=0
                        PrintGlyph(x, y, k) + PrintFrom(x, y, Dig(N, p, 13), p+13), else
PrintGlyph(x, y, k) = Glyph(x, y, Dig(k, 2, 5)/1000, Dig(k, 7, 5)/1000,  1.25^(Dig(k, 12, 1)-5), Dig(k, 0, 2))


Wa(x, y, x1, y1, x2, y2, x3, y3, x4, y4, eps) =
 Wb(min(min(x1, x2), min(x3, x4)), max(max(x1, x2), max(x3, x4)),
    min(min(y1, y2), min(y3, y4)), max(max(y1, y2), max(y3, y4)),
    x, y, x1, y1, x2, y2, x3, y3, x4, y4, eps)

Wb(minx, maxx, miny, maxy, x, y, x1, y1, x2, y2, x3, y3, x4, y4, eps)
 =    atan2((y4 - y) * (x1 - x) - (y1 - y) * (x4 - x), (x1 - x) * (x4 - x) + (y1 - y) * (y4 - y)),   if x < minx || x > maxx || y < miny || y > maxy
 =    Wc(sqrt((maxx - minx)^2 + (maxy - miny)^2), x, y, x1, y1, x2, y2, x3, y3, x4, y4, eps)

Wc(d, x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps)
 = Wa(x, y, x1, y1, (x1 + x2) / 2, (y1 + y2) / 2, (x2 + x3 + x1 + x2) / 4, (y2 + y3 + y1 + y2) / 4, (x1 + 3*x2 + 3*x3 + x4) / 8, (y4 + 3*y3 + 3*y2 + y1) / 8, eps)
   + Wa(x, y, (x4+3*x3+3*x2+x1)/8, (y4+3*y3+3*y2+y1)/8, (x4+2*x3+x2)/4, (y4+2*y3+y2)/4, (x3 + x4) / 2, (y3 + y4) / 2, x4, y4, eps),  if d > eps
 = Pi, else

Evs(x, y, k, p, eps)
  = 0, if p=0
  = Wa(x, y, Sel(k, 0), Sel(k, 1), Sel(k, 2), Sel(k, 3), Sel(k, 4), Sel(k, 5), Sel(k, 6), Sel(k, 7), eps)
    +Evs(x,y,floor(k/1000^8), p-1, eps)

Sel(a,c) =  Dig(a, c*3, 3)/1000



 */
public class SelfFormulaFastEps implements RenderJob.Boolean2dFun {

    // use static fields is optimization
    private static int L;
    private static String Nstr;
    private static int floorLog10N;

    private static double EPS;

    @Override
    public void init(List<String> args) {
        Nstr = args.get(0);
        BigInteger N = new BigInteger(Nstr); // validate
        String EPSstr = args.get(1);
        EPS = Double.parseDouble(EPSstr);
        floorLog10N = N.toString(10).length()-1; // faster variant of floor(log10(N))
        // L = Dig(N, Dig(N, 0, 2)*5 + 2, 5)
        L = Integer.parseInt(Dig(Nstr, Integer.parseInt(Dig(Nstr, 0, 2)) * 5 + 2, 5));
    }

    private static double Sr(double x, double y) {
        // Sr(x, y) = printFormula(x, y) + printN(x, y, floor(log10(N)), 0, x0, y0)
        if (y > 1.5) {
            return printFormula(x, y);
        } else if (y < 0) {
            return printN(x, y, floorLog10N, 0, 2, 0);
        }
        return printFormula(x, y) + printN(x, y, floorLog10N, 0, 2, 0);
    }

    private static double printN(double x, double y, int C, int k, double x0, double y0) {
        // printN(x, y, C, k, x0, y0) = if k>C then 0; else printN(x, y, C, k+1, x0, y0) + glyph(x, y, x0+(k mod 100)*0.4, y0-0.8*floor(k/100), 1, Dig(N,C-k,1))
        return k>C ? 0d : printN(x, y, C, k+1, x0, y0) + glyph(x, y, x0+(k % 100)*0.4, y0-0.8*(k/100), 1, Integer.parseInt(Dig(Nstr,C-k,1)));
    }

    private static double printFormula(double x, double y) {
        // printFormula(x, y) = printFrom(x, y, Dig(N, L, 13), L+13)
        return printFrom(x, y, Dig(Nstr, L, 13), L+13);
    }

    private static double printFrom(double x, double y, String k, int p) {
        // printFrom(x, y, k, p) = 0, if k=0
        //                         printGlyph(x, y, k) + printFrom(x, y, Dig(N, p, 13), p+13), else
        if (k.equals("0")) {
            return 0;
        } else {
            return printGlyph(x, y, k) + printFrom(x, y, Dig(Nstr, p, 13), p+13);
        }
    }

    private static double printGlyph(double x, double y, String k) {
        // printGlyph(x, y, k) = glyph(x, y, Dig(k, 2, 5)/1000, Dig(k, 7, 5)/1000,  1.25^(Dig(k, 12, 1)-5), Dig(k, 0, 2))
        return glyph(x, y, Double.parseDouble(Dig(k, 2, 5))/1000, Double.parseDouble(Dig(k, 7, 5))/1000,  Math.pow(1.25 , (Integer.parseInt(Dig(k, 12, 1))-5)), Integer.parseInt(Dig(k, 0, 2)));
    }

    private static double glyph(double x, double y, double x0, double y0, double s, int d) {
        // glyph(x, y, x0, y0, s, d) = Evse((x-x0)/s, (y-y0)/s, Dig(N, d*5+2, 5), Dig(N, d*5+7, 5))
        if (x-x0 > s || y-y0 > s || x < x0 || y < y0) {
            return 0;
        }
        return Evse((x-x0)/s, (y-y0)/s, Integer.parseInt(Dig(Nstr, d*5+2, 5)), Integer.parseInt(Dig(Nstr, d*5+7, 5)));
    }

    private static double Evse(double x, double y, int s, int e) {
        // Evse(x, y, s, e) = Evk(x, y, Dig(N, s, e-s), (e-s)/24)
        return Evk(x, y, Dig(Nstr, s, e-s), (e-s)/24);
    }

    private static double Evk(double x, double y, String k, int p) {
        // Evk(x, y, k, p) = Lim eps->0 |Evs(x, y, k, p, eps)|
        return Math.abs(Evs(x, y, k, p, EPS));
    }

    private static double Evs(double x, double y, String k, int p, double eps) {
        //  Evs(x, y, k, p, eps)
        //          = 0, if p=0
        //          = Wa(x, y, Sel(k, 0), Sel(k, 1), Sel(k, 2), Sel(k, 3), Sel(k, 4), Sel(k, 5), Sel(k, 6), Sel(k, 7), eps)
        //          +Evs(x,y,floor(k/1000^8), p-1, eps)
        if (p==0) {
            return 0;
        } else {
            int endIndex = k.length() - 24;
            return Wa(x, y, Sel(k, 0), Sel(k, 1), Sel(k, 2), Sel(k, 3), Sel(k, 4), Sel(k, 5), Sel(k, 6), Sel(k, 7), eps)
                      +Evs(x,y, k.substring(0, endIndex > 0 ? endIndex : 0), p-1, eps);
        }
    }

    private static double Wa(double x, double y, double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4, double eps) {
        // Wa(x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps) =
        //              Wb(min(min(x1, x2), min(x3, x4)),
        //                 max(max(x1, x2), max(x3, x4)),
        //                 min(min(y1, y2), min(y3, y4)),
        //                 max(max(y1, y2), max(y3, y4)),
        //                 x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps)

        return Wb(min(min(x1, x2), min(x3, x4)),
                  max(max(x1, x2), max(x3, x4)),
                  min(min(y1, y2), min(y3, y4)),
                  max(max(y1, y2), max(y3, y4)),
                  x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps);
    }

    private static double Wb(double minx, double maxx, double miny, double maxy, double x, double y, double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4, double eps) {
        // Wb(minx, maxx, miny, maxy, x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps)
        //    =    atan2((y4 - y) * (x1 - x) - (y1 - y) * (x4 - x), (x1 - x) * (x4 - x) + (y1 - y) * (y4 - y)),   if x < minx || x > maxx || y < miny || y > maxy
        //    =    Wc(sqrt((maxx - minx)^2 + (maxy - miny)^2), x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps)
        if (x < minx || x > maxx || y < miny || y > maxy) {
            return Math.atan2((y4 - y) * (x1 - x) - (y1 - y) * (x4 - x), (x1 - x) * (x4 - x) + (y1 - y) * (y4 - y));
        } else {
            return Wc(Math.sqrt(Math.pow(maxx - minx, 2) + Math.pow(maxy - miny, 2)), x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps);
        }
    }

    private static double Wc(double d, double x, double y, double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4, double eps) {
        // Wc(d, x,y,x1,y1,x2,y2,x3,y3,x4,y4,eps)
        //    = Wa(x, y, x1, y1, (x1 + x2) / 2, (y1 + y2) / 2, (x2 + x3 + x1 + x2) / 4, (y2 + y3 + y1 + y2) / 4, (x1 + 3*x2 + 3*x3 + x4) / 8, (y4 + 3*y3 + 3*y2 + y1) / 8, eps)
        //      + Wa(x, y, (x4+3*x3+3*x2+x1)/8, (y4+3*y3+3*y2+y1)/8, (x4+2*x3+x2)/4, (y4+2*y3+y2)/4, (x3 + x4) / 2, (y3 + y4) / 2, x4, y4, eps),  if d > eps
        //    = Pi, else
        if (d > eps) {
            return Wa(x, y, x1, y1, (x1 + x2) / 2, (y1 + y2) / 2, (x2 + x3 + x1 + x2) / 4, (y2 + y3 + y1 + y2) / 4, (x1 + 3*x2 + 3*x3 + x4) / 8, (y4 + 3*y3 + 3*y2 + y1) / 8, eps)
                    + Wa(x, y, (x4+3*x3+3*x2+x1)/8, (y4+3*y3+3*y2+y1)/8, (x4+2*x3+x2)/4, (y4+2*y3+y2)/4, (x3 + x4) / 2, (y3 + y4) / 2, x4, y4, eps);
        } else {
            return Math.PI;
        }
    }

    private static double Sel(String a, int c) {
        // Sel(a,c) =  Dig(a, c*3, 3)/1000
        return Double.parseDouble(Dig(a, c*3, 3))/1000d;
    }

    private static String Dig(String k, int p, int d) {
        // Dig(k, p, d) = floor(k/10^p) mod 10^d
        int length = k.length();
        int endIndex = length - p;
        int beginIndex = length - d - p;
        if (endIndex < 1) {
            return "0";
        }
        return k.substring(beginIndex > 0 ? beginIndex : 0, endIndex);
    }

    @Override
    public boolean fun(double x, double y) {
        return Sr(x, y) > Math.PI;
    }


}
